#!/bin/bash
# Copyright (C) 2016 - 2023 Intel Corporation
# For licensing information, see the file 'LICENSE' in the root folder

# Usage:
#   ./build_prj.sh -b build_type -g hw_gen -r rdma_core_ver -p platform -a

# defaults

BUILD_TYPE=DEV
PLATFORM=SI
GEN=HVL
ACC=0
DEFAULT_RDMA_CORE_VER=35.0

# targets passed to build.mk
BUILD=SI
LIB_RDMA_BLD_TYPE=SI

IRDMA_SRC=drivers.ethernet.rdma.linux-core-ext
RDMA_CORE= #rdma-core-35.0
LIB_IRDMA=drivers.ethernet.rdma.linux-lib-ext

function usage() {
	[ "$1" ] && echo "Error: $1"
	cat >&2 <<EOF
Usage: $0 -h -b build_type -g hw_gen -r rdma_core_ver -p platform -a

-h                print this help
-b build_type     UPSTREAM, DEV (default: DEV)
-g hw_gen         ALL - all gens
                  CVL - combined CVL/FPK (default)
                  HVL - all gen, HVL features included
-r rdma_core_ver  rdma-core release number "major.minor" or branch name (default: 44.0)
-p platform       SI, FPGA (default: SI)
-a                build for ACC (default: false)

EOF
}

############################################################
# start of script
#

if [ -f build_prj.sh ]; then
	cd ..
fi

args=`getopt b:g:r:p:ahH $*`
[ $? != 0 ] && usage
set -- $args
for arg
do
	case "$arg"
	in
	-b)
		BUILD_TYPE="$2"
		shift 2
		;;
	-r)
		RDMA_CORE_VER="$2"
		shift 2
		;;
	-p)
		PLATFORM="$2"
		shift 2
		;;
	-g)
		GEN="$2"
		shift 2
		;;
	-a)
		ACC=1
		shift
		;;
	--)
		shift
		break
		;;
	esac
done

# legacy
if [ "$#" -ne 0 ]; then
	BUILD=$1
	if [ "$1" == "SI" ]; then
		BUILD_TYPE=DEV
	elif [ "$1" == "KERNEL" ]; then
		BUILD_TYPE=UPSTREAM
	elif [ "$1" == "FPGA" ]; then
		BUILD_TYPE=DEV
		PLATFORM=FPGA
	fi
fi

if [ "$BUILD_TYPE" == "UPSTREAM" ]; then
	BUILD=KERNEL
	LIB_RDMA_BLD_TYPE=UPSTREAM

	# validate parameters for upstream build
	if [ "$PLATFORM" != "SI" -o "$GEN" == "HVL" -o $ACC -eq 1 ]; then
		echo "invalid arguments"
		exit 1
	fi
else
	if [ "$PLATFORM" == "FPGA" ]; then
		BUILD=FPGA
		LIB_RDMA_BLD_TYPE=FPGA
		BUILD_EXT="_fpga"
	fi

	if [ "$GEN" == "HVL" ]; then
		DEFAULT_RDMA_CORE_VER=35.0
	fi

	if [ -z $RDMA_CORE_VER ]; then
		RDMA_CORE_VER=$DEFAULT_RDMA_CORE_VER
	fi

	LIB_RDMA_PREP_PATCH="irdma-$RDMA_CORE_VER-prep.patch"
	IWPM_PATCH="iwpmd-$RDMA_CORE_VER-prep.patch"

	if [ "$GEN" == "HVL" ]; then
		# all gen builds with HVL specific content
		if [ "$ACC" -eq 1 ]; then
			BUILD=ALL_GEN_HVL_ACC
			LIB_RDMA_BLD_TYPE=ALL_GEN_HVL_ACC
			LIB_RDMA_PREP_PATCH="irdma-mev-acc-$RDMA_CORE_VER-prep.patch"
			BUILD_EXT='_acc'
		else
			BUILD=ALL_GEN_HVL
			LIB_RDMA_BLD_TYPE=ALL_GEN_HVL
			LIB_RDMA_PREP_PATCH="irdma-mev-$RDMA_CORE_VER-prep.patch"
		fi
		BUILD_EXT+="-hvl"
	elif [ "$GEN" == "ALL" ]; then
		BUILD=ALL_GEN
		LIB_RDMA_BLD_TYPE=ALL_GEN
	fi
fi

echo "BUILD_TYPE: $BUILD_TYPE"
echo "PLATFORM: $PLATFORM"
echo "GEN: $GEN"
echo "ACC: $ACC"
echo "RDMA_CORE_VER: $RDMA_CORE_VER"
echo
echo "DRV_BUILD: $BUILD"
echo "LIB_BUILD: $LIB_RDMA_BLD_TYPE"
echo

RDMA_CORE_DIR="rdma-core-$RDMA_CORE_VER"
RDMA_CORE_TAR="rdma-core-$RDMA_CORE_VER.tar.gz"
RDMA_CORE_ZIP="rdma-core-$RDMA_CORE_VER.tgz"

PF_VERSION=`make -s -C $IRDMA_SRC/src -f build.mk version`

TARGET_DIR=irdma-build

IRDMA_DEST="$TARGET_DIR/irdma-$PF_VERSION$BUILD_EXT/src/irdma"
SPDX="GPL-2.0 or Linux-OpenIB"

mkdir -p $IRDMA_DEST

IRDMA_TAR="irdma-$PF_VERSION$BUILD_EXT.tgz"

if [ "$BUILD_TYPE" != "UPSTREAM" ]; then
	# SI build currently tracks OOT combined driver supporting FPK, CVL

	# Untar lan src and grab required .h file
	#I40E_TAR=`find . -maxdepth 1 -name "i40e-*.tar.gz" | head -1`
	#if [ -n "$I40E_TAR" ]; then
	#	echo "I40E: $I40E_TAR"
	#	tar xzf $I40E_TAR
	#	DEVID_DIR=`find . -maxdepth 1 -name "i40e-*" -type d | head -1`
	#	cp $DEVID_DIR/src/i40e_client.h $IRDMA_DEST/
	#	sed -i "s/GPL-2.0/$SPDX/g" $IRDMA_DEST/i40e_client.h
	#else
	#	echo "ERROR: Cannot find source tarball for I40E"
	#	echo
	#	exit 1
	#fi

	if [ "$GEN" == "HVL" -o "$GEN" == "ALL" ]; then
		# setup idpf headers
		# Add EXTRA_INC=<path_to_idpf_src_code>
		EXTRA_INC=$(pwd)/drivers.ethernet.linux.mev-dp/silicon/idpf_0.0.625/idpf/src/
		IDPF_TAR=`find .  -maxdepth 3  -name "idpf_*.tar.gz" | head -1`
		echo EXTRA_INC is $EXTRA_INC
		echo IDPF_TAR is $IDPF_TAR
		if [ -n "$IDPF_TAR" ]; then
			tar xvzf $IDPF_TAR
                        IDPF_DIR=`find . -maxdepth 3 -name "idpf_*" -type d | head -1`
			DEVID_DIR=$IDPF_DIR/idpf
			echo IDPF_DIR is $IDPF_DIR
			echo DEVID_DIR is $DEVID_DIR
			cp $IDPF_DIR/idpf/src/iidc.h $IRDMA_DEST/
			sed -i "s/GPL-2.0/$SPDX/g" $IRDMA_DEST/iidc.h
		else
			echo "ERROR: Cannot find source tarball for IDPF"
			echo
			exit 1
		fi
	fi
fi

echo

# build/strip irdma source
cd $IRDMA_SRC/src
make -f build.mk clean all BUILD=$BUILD
cp -rf irdma ../../$IRDMA_DEST/..
cd ../..

# unzip rdma-core and apply prep patch
#cd $RDMA_CORE
#rm -rf $RDMA_CORE_DIR
#rm -rf rdma-core-copy
#
#if [ ! -f "${RDMA_CORE_TAR}" ]; then
#	# Check if rdma-core version is in major.minor format.
#	# For simplicity, assume it consists of two numbers separated by '.'.
#	# Otherwise (no '.' in version string) it is a branch name.
#	a=( ${RDMA_CORE_VER//./ } )
#	if [ -n "${a[1]}" ]; then
#		# major.minor
#		wget https://github.com/linux-rdma/rdma-core/archive/refs/tags/v${RDMA_CORE_VER}.tar.gz \
#			-O ${RDMA_CORE_TAR}
#	else
#		# branch name
#		wget https://github.com/linux-rdma/rdma-core/archive/refs/heads/${RDMA_CORE_VER}.tar.gz \
#			-O ${RDMA_CORE_TAR}
#	fi
#	if [ $? -ne 0 ]; then
#		echo "ERROR: Failed to download source tarball for rdma-core $RDMA_CORE_VER"
#		echo
#		exit 1
#	fi
#fi
#
#tar -xzf $RDMA_CORE_TAR
#cp -r $RDMA_CORE_DIR rdma-core-copy
#cd $RDMA_CORE_DIR
#
## Apply prep- patch if exists.
## The rdma-core >= 37.0 does not require a prep- patch, as it already
## has irdma provider.
## For rdma-core < 37.0, we only have prep- patches for versions
## 22.1, 27.0 and 35.0.  Exit if prep patch is missing for given version.
#
#if [ ! -d providers/irdma ]; then
#	if [ ! -f ../../$LIB_IRDMA/prep-$RDMA_CORE_VER/$LIB_RDMA_PREP_PATCH ]; then
#		echo "ERROR: irdma-prep patch is missing for rdma-core $RDMA_CORE_VER"
#		echo
#		exit 1
#	fi
#
#	patch -p1 < ../../$LIB_IRDMA/prep-$RDMA_CORE_VER/$LIB_RDMA_PREP_PATCH
#	ret=$?
#	if [ $ret -ne 0 ]; then
#		echo "build error: prep patch did not apply"
#		exit 1
#	fi
#	mkdir -p providers/irdma
#	cp ../../$LIB_IRDMA/prep-$RDMA_CORE_VER/CMakeLists.txt providers/irdma/CMakeLists.txt
#	if [ "$GEN" != "HVL" ]; then
#		if [ ! -f ../../$LIB_IRDMA/prep-$RDMA_CORE_VER/$IWPM_PATCH ]; then
#			echo "ERROR: iwpmd-prep patch is missing for rdma-core $RDMA_CORE_VER"
#			echo
#			exit 1
#		else
#			patch -p1 < ../../$LIB_IRDMA/prep-$RDMA_CORE_VER/$IWPM_PATCH
#			ret=$?
#			if [ $ret -ne 0 ]; then
#				echo "build error: prep patch did not apply"
#				exit 1
#			fi
#		fi
#	fi
## For rdma-core 42.0 or later apply a prep patch to enable iwpmd service for irdma in iWARP mode.
## The changes in the prep patch have been submitted upstream for review and when
## accepted the prep patch won't be necessary for future rdma-core versions.
#elif [ `echo "$RDMA_CORE_VER >= 42.0" | bc` -eq 1 ]; then
#	if [ "$GEN" != "HVL" ]; then
#		if [ ! -f ../../$LIB_IRDMA/prep-$RDMA_CORE_VER/$IWPM_PATCH ]; then
#			echo "ERROR: iwpmd-prep patch is missing for rdma-core $RDMA_CORE_VER"
#			echo
#			exit 1
#		else
#			patch -p1 < ../../$LIB_IRDMA/prep-$RDMA_CORE_VER/$IWPM_PATCH
#			ret=$?
#			if [ $ret -ne 0 ]; then
#				echo "build error: prep patch did not apply"
#				exit 1
#			fi
#		fi
#	fi
#fi
#cd ../../
#
## build userspace and copy over to rdma-core
#if [ ! -d $LIB_IRDMA/src/DRIVER_CORE ]; then
#	cp -r -f $IRDMA_SRC $LIB_IRDMA/src/DRIVER_CORE
#else
#	cp -r -f $IRDMA_SRC/* $LIB_IRDMA/src/DRIVER_CORE/
#fi
#cd $LIB_IRDMA/src
#make -f build.mk clean all BUILD=$LIB_RDMA_BLD_TYPE
#cp -r -f libirdma/src/* ../../$RDMA_CORE/$RDMA_CORE_DIR/providers/irdma/
#rm -f ../../$RDMA_CORE/$RDMA_CORE_DIR/providers/irdma/irdma-abi.h
#cp -r -f libirdma/src/irdma-abi.h ../../$RDMA_CORE/$RDMA_CORE_DIR/kernel-headers/rdma/irdma-abi.h
#echo "ls kernel-headers"
#ls -al ../../$RDMA_CORE/$RDMA_CORE_DIR/kernel-headers/rdma
#echo "ls libirdma/src"
#ls -al libirdma/src
#cd ../..
#
## rezip modified rdma-core
#cd $RDMA_CORE
#tar -czf $RDMA_CORE_ZIP $RDMA_CORE_DIR
#cd ../
#
## generate new rdma-code patch for userspace
#diff -N -u -r -x '.clang-format' -x '.git*' -x '.tr*' -x '.ma*' -x 'pandoc*' $RDMA_CORE/rdma-core-copy $RDMA_CORE/$RDMA_CORE_DIR > libirdma-$RDMA_CORE_VER.patch
#rm -rf $RDMA_CORE/rdma-core-copy

# finish staging top level directory
cp $IRDMA_SRC/README_irdma.txt $TARGET_DIR/..
cp $IRDMA_SRC/readmefirst.txt $TARGET_DIR/..

# finish staging irdma
if [ "$GEN" != "HVL" ]; then
	cp $IRDMA_SRC/COPYING $IRDMA_DEST/../..
	cp $IRDMA_SRC/README_irdma.txt $IRDMA_DEST/../..
	cp $IRDMA_SRC/readmefirst.txt $IRDMA_DEST/../..
	cp libirdma-$RDMA_CORE_VER.patch $IRDMA_DEST/../..
	if [ -f $LIB_IRDMA/build_core.sh -a -f $LIB_IRDMA/install_core.sh ]; then
		cp $LIB_IRDMA/build_core.sh $LIB_IRDMA/install_core.sh $IRDMA_DEST/../..
		chmod +x $IRDMA_DEST/../../build_core.sh $IRDMA_DEST/../../install_core.sh
	fi

	if [ -z "${COPYRIGHT_S}" ]; then
		COPYRIGHT_S=/dev/null
	fi

#	gawk -v lid="$(cat $COPYRIGHT_S)" 'BEGIN{lidadded=0}
#	{if (substr($0,1,2) == "#!") {print $0; print lid; lidadded=1;} else {if (lidadded==0) {lid=sprintf("%s\n%s",lid,$0);} else {print $0;}}}
#' $IRDMA_DEST/build.sh > $IRDMA_DEST/../../build.sh
        cp $IRDMA_DEST/build.sh $IRDMA_DEST/../../build.sh
else
	cp $IRDMA_DEST/build.sh $IRDMA_DEST/../../build.sh
fi

chmod +x $IRDMA_DEST/../../build.sh
rm -f $IRDMA_DEST/build.sh

cd $TARGET_DIR
tar -c --owner=0 --group=0 -zf $IRDMA_TAR irdma-$PF_VERSION$BUILD_EXT

mv *tgz ..

## if called from Jenkins, skip build step
#if [ "$JENKINS" == "true" ]; then
#	exit 0
#fi
#
## RPM GENERATION
#if [ "$SANDBOX_DIR" != "" -a "$GEN" != "HVL" ]; then
#	# Put the auxbus header temporarily in Sandbox:
#	cp $SANDBOX_DIR/$DEVID_DIR/src/linux/auxiliary_bus.h $SANDBOX_DIR
#	cp $SANDBOX_DIR/$IRDMA_SRC/intel-public-key-irdma-ko.rsa $SANDBOX_DIR
#
#	# call the build_rpm.sh and relocate to TARGET_DIR after done:
#	source ${SANDBOX_DIR}/${IRDMA_SRC}/build_rpm.sh
#	cd ${SANDBOX_DIR}/${TARGET_DIR}
#fi

pushd irdma-$PF_VERSION$BUILD_EXT
./build.sh noinstall $EXTRA_INC
ret=$?
if [ $ret -ne 0 ]; then
	echo "build error"
	exit 1
fi
popd

## build rdma-core
#pushd ../$RDMA_CORE/$RDMA_CORE_DIR
#./build.sh
#popd
