#!/bin/bash
# Copyright (C) 2022 Intel Corporation
# For licensing information, see the file 'LICENSE' in the root folder

if [ "$RPMRELEASE" == "" ]; then
	RPMRELEASE="1"
fi

# SANDBOX_DIR should be populated in Component Builds:
if [ "$SANDBOX_DIR" != "" ]; then
	echo "Processing buildrpm.sh..."

	# If VERSION_ID is not populated, load information about OS-Release:
	if [ "${VERSION_ID}" == "" ]; then
		if [ -f /etc/os-release ]; then
			# Pick the component image distro information:
			echo "Load os-release from component tool image"
			source /etc/os-release
		fi
	fi

	KERNELVER=`uname -r | gawk 'BEGIN{FS="-"}{print $1}'`
	FULLKERNELVER=`uname -r`
	BUILDRPM=0
	ARCH=`uname -m`
	echo "Architecture: ${ARCH}"

	# In sequential builds, the resultant files are accumulated in ${SANDBOX_DIR}/intermediate_resultants/
	# Fedora is expected to be the first one to run and succeed.
	# To use the rpmbuild -t<x> option, unpack the tarball, inject the spec file and re-pack the tarball.

	TARBALL=`find ${SANDBOX_DIR}/intermediate_resultants/ -name "irdma*.tgz"`
	mkdir ${SANDBOX_DIR}/tmptar

	# Copy the initial tarball to its final location
	cp ${TARBALL} ${SANDBOX_DIR}

	# Copy tarball to temporary location to inject spec file:
	cp ${TARBALL} ${SANDBOX_DIR}/tmptar
	cd ${SANDBOX_DIR}/tmptar
	tar zxvf irdma*.tgz
	cp ${SANDBOX_DIR}/${IRDMA_SRC}/irdma.spec $(find . -name src)
	cp ${SANDBOX_DIR}/${IRDMA_SRC}/sign-file $(find . -name src)
	tar -c --owner=0 --group=0 -vzf $IRDMA_TAR irdma-$PF_VERSION$BUILD_EXT

	echo "TARBALL: ${TARBALL}  SPECTAR: ${SPECTARBALL}"

	if [ "${ID}" == "rhel" ]; then
		# RHEL Specific sufix for RPM
		RPMNAMEPREFIX="kmod-irdma-"
		RPMNAMESUFFIX="-${RPMRELEASE}.${ID}${VERSION_ID}"
		BUILDRPM=1

		if [ "${VERSION_ID}" == "8.7" ]; then
			# RHEL 9.1 does not have an issue with kernel-doc
			# -o "${VERSION_ID}" == "9.1" ]; then
			echo "RHEL ${VERSION_ID} Beta bug handling"
			grep -n members /usr/src/kernels/$(uname -r)/scripts/kernel-doc | grep DECLARE_FLEX_ARRAY
			cat  /usr/src/kernels/$(uname -r)/scripts/kernel-doc | gawk -F  ' |\t|[?]|\\\\' '{
				id=0;
				for (i=1; i<=NF; i++) {
					if ($i == "$members") id++;
					if ($i == "DECLARE_FLEX_ARRAY") id++;
				}
				if (id ==2) {
					print "\t$members =~ s/(?:__)?DECLARE_FLEX_ARRAY\\s*\\(([^,)]+),\\s*([^,)]+)\\)/$1 $2\\[\\]/gos;"
				} else {
					print $0
				} }' > $SANDBOX_DIR/kernel-doc
			DIF=`diff /usr/src/kernels/$(uname -r)/scripts/kernel-doc $SANDBOX_DIR/kernel-doc | wc -l`
			echo "${DIF} differences found in kdoc"
			cp -f $SANDBOX_DIR/kernel-doc /usr/src/kernels/$(uname -r)/scripts/kernel-doc
		fi

		# Kernels for RHEL are under /usr/src/kernels/
		KERNELSRC="/usr/src/kernels/$(uname -r)"
		KERNELINC="${KERNELSRC}/include/linux"
		AUX_BUS="${KERNELINC}/auxiliary_bus.h"
	fi

	if [ "${ID}" == "sles" ]; then
		# SLES Specific sufix for RPM
		RPMNAMEPREFIX="irdma-kmp-default-"
		SLESKERNELVER=`echo ${FULLKERNELVER} | gawk -F '-' '{print $1"_"$2}'`
		RPMNAMESUFFIX="_k${SLESKERNELVER}-${RPMRELEASE}.${ID}${VERSION_ID}"
		BUILDRPM=1

		# Kernels for SLES are under /usr/src
		KERNELSRC="/usr/src/linux"
		KERNELINC="${KERNELSRC}/include/linux"
		AUX_BUS="${KERNELINC}/auxiliary_bus.h"
	fi

	if [ "${BUILDRPM}" == "1" ]; then
		SPECLOC="$SANDBOX_DIR/tmptar"

		# OLE Team indicated some instability on the kerberos authentication system
		# Which results on the following signing error:
		#    Could not retrieve oAuthToken using Curl. Please make sure curl is installed
		#    and computer is AD linked.Unable to connect to server https://edss-api.intel.com:
		#    SOAP initialization failed: For authentication curl --negotiate did not work,
		#    so you may need to provide either user name and password or oAuth token
		# See HSD: https://hsdes.intel.com/resource/22016132211
		# Adding call to restart the service and update tokens ahead of the signing
		systemctl restart kerberos

		echo "Moving to ${SPECLOC}"
		cd ${SPECLOC}

		DIF=`diff ${SANDBOX_DIR}/auxiliary_bus.h ${AUX_BUS} | wc -l`
		echo "${DIF} differences found in aux bus"

		if [ ! -f ${AUX_BUS} ]; then
			echo "aux bus not found. Copying ice version"
			cp -f ${SANDBOX_DIR}/auxiliary_bus.h ${AUX_BUS}
		fi

		RPMVER=`rpm --version`

		echo "********************************************"
		echo "rpmbuild v.${RPMVER}"
		echo "********************************************"

		DISTROROOT="${SANDBOX}/${ID}/${VERSION_ID}"
		RPMBUILDDIR="${SANDBOX_DIR}/${ID}/${VERSION_ID}/rpmbuild"
		RPMFILENAME="${RPMNAMEPREFIX}${PF_VERSION}${RPMNAMESUFFIX}"

		echo "Create $RPMNAMEPREFIX$PF_VERSION$RPMNAMESUFFIX.${ARCH}.rpm ..."
		# rpmbuild into a OS Distro specific location:
		rpmbuild -ta irdma-${PF_VERSION}.tgz --define '_IRDMA_VER_ '${PF_VERSION} --define '_KERNELSRC_ '${KERNELSRC} --define '_topdir '${RPMBUILDDIR} -vv

		echo "Copy $RPMNAMEPREFIX$PF_VERSION$RPMNAMESUFFIX.${ARCH}.rpm ..."
		# Pick the rpm files, copy and rename:
		if [ -f ${RPMBUILDDIR}/RPMS/${ARCH}/irdma-${PF_VERSION}-1.${ARCH}.rpm ]; then
			FULLRPMFILENAME="${SANDBOX_DIR}/${RPMFILENAME}.${ARCH}.rpm"
			cp ${RPMBUILDDIR}/RPMS/${ARCH}/irdma-${PF_VERSION}-1.${ARCH}.rpm  ${FULLRPMFILENAME}

			# Sign RPM:
			SignFile ${FULLRPMFILENAME} -vv -dc -l log.txt -c Linux_RPM_GPG_Release_SO
		else
			echo "RPM file not found: ${RPMBUILDDIR}/RPMS/x86_64/irdma-${PF_VERSION}-1.x86_64.rpm"
			ls -lart ${RPMBUILDDIR}/RPMS/x86_64/
		fi
		if [ -f ${RPMBUILDDIR}/SRPMS/irdma-${PF_VERSION}-1.src.rpm ]; then
			cp ${RPMBUILDDIR}/SRPMS/irdma-${PF_VERSION}-1.src.rpm ${SANDBOX_DIR}/${RPMFILENAME}.src.rpm
		else
			echo "SRPM file not found: ${RPMBUILDDIR}/SRPMS/irdma-${PF_VERSION}-1.src.rpm"
			ls -lart ${RPMBUILDDIR}/SRPMS/
		fi
	fi
else
	echo "SANDBOX_DIR is empty"
fi

