#!/usr/bin/perl
# SPDX-License-Identifier: GPL-2.0-only
# Copyright (C) 2014 - 2019 Intel Corporation

=head1 NAME

cppkeep.pl - keep or strip CPP #ifdef flags, and warn about unknown flags

=head1 SYNOPSIS

cppkeep.pl [--stderr] <unifdef> [keep symbol] .. -[strip symbol] ..

  Options:
    --missing=<FILE> - output unknown flags to <FILE>
    --stderr         - output unknown flags to stderr instead of to a file

  Arguments:
    <unifdef>      - path to valid unifdef stripping program
    [keep symbol]  - symbol you wish to keep in outputted source
    [strip symbol] - symbol you wish to remove from outputted source

  Examples:

    cppkeep.pl ./scripts/unifdef SOME_FLAG -OTHER_FLAG KEEP_THIS_FLAG .. <input >output

    cppkeep.pl will output an unknown CPP macros (those not listed in the set
    of flags provided as arguments) to THESE_UNKNOWN_FLAGS_NEED_PLACING_IN_BUILDMK.
    This is for compatibility with the original shell script implementation. If desired,
    you may pass --missing=<FILE> to output unknown flags to a different file
    of your choice, or you may pass --stderr in order to output missing flags
    to stderr instead.

=head1 DESCRIPTION

This program is designed to keep or strip certain #ifdef flags from C source
code, and additionally output a list of C preprocessor macros that were not
specified. This is used by build setups in order to easily keep track of build
macros and to help simplify C source code by removing #ifdef logic prior to
shipping code.

The program takes advantage of unifdef as the C preprocessor tool, and uses
this to determine a complete liste of macros used in conditional expressions
found in the source code (provided on stdin).

If given command line parameters it will keep or strip the #ifdef and #if C
Pre Processor conditional expressions, in a manner similar to unifdef.

Any macros not provided as a keep or strip symbol will be stored by default in
THESE_UNKNOWN_FLAGS_NEED_PLACING_IN_BUILD_MK for compatibility as a drop in
replacement for the original implementation.

To avoid races, when outputting to a file, it will take the current contents
of the missing flags file and sort them along with new output before
overwriting the old contents. This ensures that multiple instances of
cppkeep.pl can safely output to the same file without losing contents.

If you use --stderr output, this behavior is avoided and flags are output
directly to stderr as they are found.

=head1 AUTHOR

Jacob Keller <jacob.e.keller@intel.com>

=cut

use strict;
use warnings;

use Getopt::Long qw(GetOptions);
use Pod::Usage qw(pod2usage);
use File::Basename qw(fileparse);
use File::Temp qw(tempfile);
use File::Copy qw(copy);

my $progname = fileparse($0);

# Find unique elements of array, preserving order
sub uniq {
    my %seen;
    return grep { !$seen{$_}++ } @_;
}

# Get the options
my %options = ();
$options{"missing"} = "THESE_UNKNOWN_FLAGS_NEED_PLACING_IN_BUILD_MK";
Getopt::Long::Configure("default", "pass_through", "auto_help");
GetOptions(\%options, 'stderr!', "missing=s");

exists $ARGV[0] or pod2usage(1);

# Get the arguments
my ($unifdef_prog, @args) = @ARGV;

# A hash of known symbols indicating whether to keep or strip them
my %known_symbols;

foreach (@args) {
    # Note that substitution occurs inplace for why this works
    $known_symbols{$_} = s/^-// ? 1 : 0;
}

# Make a copy of the standard input for use
my ($source_fh, $source) = tempfile(UNLINK => 1);
copy(\*STDIN,$source);

# Make a temporary file to store the discovered symbols
my ($unifdef_symbols_fh, $unifdef_symbols) = tempfile(UNLINK => 1);

# Process the source input with unifdef to find all known symbols, using the -s option
system("$unifdef_prog -s $source >$unifdef_symbols");

# Find a list of all the unknown symbols used in the standard input
my @found_unique = uniq(<$unifdef_symbols_fh>);
chomp @found_unique;

# Exclude any symbols specified on the input
my @unknown_symbols = grep {!exists($known_symbols{$_}) } @found_unique;

# Generate list of arguments for symbols which we want to remove the code and
# #if logic
my @strip_flags = map("-U" . $_, @unknown_symbols);
my $strips = join(' ', @strip_flags);

# Generate list of arguments for symbols which we want to remove #if logic, but
# leave code
my @keep_symbols = grep { $known_symbols{$_} } keys %known_symbols;
my @keep_flags = map("-D" . $_, @keep_symbols);
my $keeps = join(' ', @keep_flags);

# Finally, run unifdef again with the strip and keep flags required
system("$unifdef_prog $strips $keeps $source");

if ( $#unknown_symbols >= 0) {
    my ($fh, $unknowns) = tempfile(UNLINK => 1);
    foreach (@unknown_symbols) { print $fh "$_\n"; }

    if ($options{'stderr'}) {
        system("sort $unknowns -u >&2");
    } else {
        my $results = $options{"missing"};

        # If the output file exists, consume its input and sort it into the
        # results we found for the current run. This is for compatibility with
        # the shell script implementation. Note that this works because sort
        # needs all input before processing, so we just overwrite the resulting
        # file.
        if (-f $results) {
            system("sort $unknowns $results -u -o $results");
        } else {
            system("sort $unknowns -u -o $results");
        }
    }
}
