#!/usr/bin/python3
#
# Copyright (c) 2021 Intel Corporation
#
# This script adds (or updates) comments in lines with #else or #endif,
# so it makes it easier to identify matching #if, #elif, #ifdef or #ifndef
# directives. E.g.
#
#     BEFORE                   |      AFTER
#     -------------------------------------------------
#     #ifdef XXX               |      #ifdef XXX
#     ...                      |      ...
#     #elif YYY                |      #elif YYY
#     ...                      |      ...
#     #else                    |      #else /* YYY */
#     ...                      |      ...
#     #endif                   |      #endif /* YYY */
#
# It is intended to be used with C or C++ source files, but it should
# work with any source code that uses C-style preprocessor syntax (e.g. C#).
#
# If no filename(s) is specified in command-line args, the script reads from
# standard input.  Otherwise, it processes all the files specified
# in command-line arguments (wildcards allowed).
# By default output goes to stdout, unless '-i' flag is specified.
#
# Command-line options:
# -c, --cpp      - enable C++ comments ('//' instead of '/* ... */')
# -t, --tab      - indent comments with single TAB instead of space
# -n, --notdef   - add '!' in case of #ifndef condition, e.g. '#else /* !XXX */'
# -i, --inplace  - edit files in place
# -p, --preserve - preserve indentation
# -h, --help     - print this help
#
# Examples:
#   $ ifdef.py -i *.c
#   $ ifdef.py -c pre.cpp > post.cpp
#   $ cat pre.c | ifdef.py -t > post.cpp
#

import sys
import getopt
import re
import fileinput

# defaults
Inplace = False
Notdef = False
Preserve = False
C_start = '/* '
C_end = ' */'
C_indent = ' '

p_if = re.compile('^\s*#if\S*\s*')
p_ifndef = re.compile('^\s*#ifndef\s*')
p_elif = re.compile('^\s*#elif\s*')
p_else = re.compile('^(\s*)#else\s*')
p_endif = re.compile('^(\s*)#endif\s*')

def usage(prog):
    print(f'usage: {prog} [-ctnih] [filename ...]')
    print('\t-c, --cpp      - enable C++ comments')
    print('\t-t, --tab      - indent comments with TAB instead of space')
    print('\t-n, --notdef   - add "!" in case of #ifndef condition')
    print('\t-i, --inplace  - edit files in place')
    print('\t-p, --preserve - preserve indentation')
    print('\t-h, --help     - print this help')

def process_file(filename):
    cond = ""
    stack = []

    for line in fileinput.input(filename, inplace=Inplace):
        m = p_if.match(line)
        if (m):
            cond = line[m.end():-1]
            if (Notdef):
                m = p_ifndef.match(line)
                if (m):
                    cond = '!' + cond
            for c in stack:
                if cond in c:
                    print(f'{fileinput.filename()}:{fileinput.filelineno()}: nested cond: {cond}', file=sys.stderr)

            stack.append(cond)
            if (Preserve):
                print(f'{line.rstrip()}')
            else:
                print(f'{line.strip()}')
            continue

        m = p_else.match(line)
        if (m):
            if (Preserve):
                print(f'{m.group(0).rstrip()}{C_indent}{C_start}{cond}{C_end}')
            else:
                print(f'{m.group(0).strip()}{C_indent}{C_start}{cond}{C_end}')
            continue

        m = p_elif.match(line)
        if (m):
            cond = line[m.end():-1]
            stack.pop()
            stack.append(cond)
            print(f'{line}', end='')
            continue

        m = p_endif.match(line)
        if (m):
            if (Preserve):
                print(f'{m.group(0).rstrip()}{C_indent}{C_start}{cond}{C_end}')
            else:
                print(f'{m.group(0).strip()}{C_indent}{C_start}{cond}{C_end}')
            stack.pop()
            if (stack):
                cond = stack[-1]
            else:
                cond = ''
            continue

        print(f'{line.rstrip()}')

    print(f'processed file: {filename}', file=sys.stderr)


# main

try:
    opts, args = getopt.getopt(sys.argv[1:], 'hcnitp',
            [ 'help', 'cpp', 'notdef', 'inplace', 'tab', 'preserve' ])
except getopt.GetoptError as e:
    print(f'error: {e}', file=sys.stderr)
    usage(sys.argv[0])
    sys.exit(1)

for opt, arg in opts:
    if opt in ('-h', '--help'):
        usage(sys.argv[0])
        sys.exit(0)
    if opt in ('-c', '--cpp'):
        C_start = '// '
        C_end = ''
    elif opt in ('-n', '--notdef'):
        Notdef = True
    elif opt in ('-i', '--inplace'):
        Inplace = True
    elif opt in ('-t', '--tab'):
        C_indent = '\t'
    elif opt in ('-p', '--preserve'):
        Preserve = True
    else:
        print(f'invalid option: {atgv}', file=sys.stderr)
        usage(sys.srgv[0])
        sys.exit(1)

if (len(args) > 0):
    for filename in args:
        try:
            process_file(filename)
        except:
            print(f'cannot process file: {filename}', file=sys.stderr)
else:
    process_file('-')
