/* Copyright (c) 2016 - 2023 Intel Corporation */
#ifndef LICENSE_REPLACED
/* For licensing information, see the file 'LICENSE' in the root folder */
#endif /* LICENSE_REPLACED */
#include <linux/netdevice.h>
#include <net/genetlink.h>
#include <net/netlink.h>
#include <asm/delay.h>

#ifndef HPM_LINUX
#include "status.h"
#endif
#include "osdep.h"
#include "type.h"

#include "main.h"
#include <irdma_register.h>

#include "dbg.h"

static LIST_HEAD(icrdmadbg_iw_dev_list);
static LIST_HEAD(icrdmadbg_evt_ctx_wait_list);

static DEFINE_SPINLOCK(icrdmadbg_iw_dev_list_lock);
static DEFINE_SPINLOCK(icrdmadbg_evt_ctx_wait_list_lock);
/*
 * netlink family responsilble for communication with user space debug tools
 */
static struct genl_family icrdmadbg_nl_family = {
	.id = GENL_ID_GENERATE,
	.name = IRDMADBG_GENL_NAME,
	.version = IRDMADBG_GENL_VERSION,
	.maxattr = IRDMADBG_ATTR_MAX,
	.netnsok = true,
};

static struct genl_multicast_group icrdmadbg_events_mc_group = {
	.name = IRDMADBG_GENL_EVENTS_MC_GROUP_NAME,
};

static struct nla_policy icrdmadbg_nl_policy[IRDMADBG_ATTR_COUNT] = {
	[IRDMADBG_ATTR_UNSPEC] = {.type = NLA_UNSPEC},
	[IRDMADBG_ATTR_IFINDEX] = {.type = NLA_U32},
	[IRDMADBG_ATTR_REG] = {.type = NLA_NESTED},
	[IRDMADBG_ATTR_HMC] = {.type = NLA_NESTED},
	[IRDMADBG_ATTR_EVENT_CFG] = {.type = NLA_NESTED},
	[IRDMADBG_ATTR_EVENT] = {.type = NLA_NESTED},
};

static struct nla_policy icrdmadbg_nl_reg_policy[IRDMADBG_ATTR_REG_COUNT] = {
	[IRDMADBG_ATTR_REG_UNSPEC] = {.type = NLA_UNSPEC},
	[IRDMADBG_ATTR_REG_OP_TYPE] = {.type = NLA_U8},
	[IRDMADBG_ATTR_REG_ADDR] = {.type = NLA_U32},
	[IRDMADBG_ATTR_REG_VALUE] = {.type = NLA_U32},
};

static struct nla_policy icrdmadbg_nl_hmc_policy[IRDMADBG_ATTR_HMC_COUNT] = {
	[IRDMADBG_ATTR_HMC_UNSPEC] = {.type = NLA_UNSPEC},
	[IRDMADBG_ATTR_HMC_OBJ_TYPE] = {.type = NLA_U32},
	[IRDMADBG_ATTR_HMC_OBJ_ID_REQ] = {.type = NLA_U32},
	[IRDMADBG_ATTR_HMC_VALID_ONLY] = {.type = NLA_FLAG},
	[IRDMADBG_ATTR_HMC_QP_NO_CQP] = {.type = NLA_FLAG},
	[IRDMADBG_ATTR_HMC_QP_FREEZE] = {.type = NLA_FLAG},
	[IRDMADBG_ATTR_HMC_QP_RAW] = {.type = NLA_FLAG},
	[IRDMADBG_ATTR_HMC_OBJ_SIZE] = {.type = NLA_U32},
	[IRDMADBG_ATTR_HMC_OBJ_CNT] = {.type = NLA_U32},
	[IRDMADBG_ATTR_HMC_OBJ_START_ID] = {.type = NLA_U32},
	[IRDMADBG_ATTR_HMC_OBJ_DATA] = {.type = NLA_BINARY},
};

static struct nla_policy
    icrdmadbg_nl_event_cfg_policy[IRDMADBG_ATTR_EVENT_CFG_COUNT] = {
	[IRDMADBG_ATTR_REG_UNSPEC] = {.type = NLA_UNSPEC},
	[IRDMADBG_ATTR_EVENT_CFG_MONITOR_STATE] = {.type = NLA_U32},
	[IRDMADBG_ATTR_EVENT_CFG_LIST_EVT_TYPE] = {.type = NLA_NESTED},
	[IRDMADBG_ATTR_EVENT_CFG_QUEUE_TYPE] = {.type = NLA_U32},
	[IRDMADBG_ATTR_EVENT_CFG_QUEUE_ID] = {.type = NLA_U32},
};

#ifdef IW_DEBUG_PRINT
/*
 * temp function for resources status printout
 */
static void print_resources_status(struct icrdmadbg_hmc_dump_ctx *dump_ctx)
{
	int k, l, m, status;

	printk(KERN_INFO "%s,%d === QPs ===\n", __func__, __LINE__);

	for (k = 0; k < 8; k++) {
		for (l = 0;
		     l <
		     (dump_ctx->iwdev->sc_dev.hmc_info->hmc_obj[I40E_HMC_IW_QP].
		      cnt / 8); l++) {
			m = k * 64 + l;
			status =
			    irdma_is_resource_allocated(dump_ctx->iwdev,
							dump_ctx->iwdev->
							allocated_qps,
							m) ? 1 : 0;
			printk("%d ", status);
		}
		printk("\n");
	}
	printk(KERN_INFO "%s,%d === CQs ===\n", __func__, __LINE__);

	for (k = 0; k < 8; k++) {
		for (l = 0;
		     l <
		     (dump_ctx->iwdev->sc_dev.hmc_info->hmc_obj[I40E_HMC_IW_CQ].
		      cnt / 8); l++) {
			m = k * 64 + l;
			status =
			    irdma_is_resource_allocated(dump_ctx->iwdev,
							dump_ctx->iwdev->
							allocated_cqs,
							m) ? 1 : 0;
			printk("%d ", status);
		}
		printk("\n");
	}
	printk(KERN_INFO "%s,%d === MRs ===\n", __func__, __LINE__);

	for (k = 0; k < 8; k++) {
		for (l = 0;
		     l <
		     (dump_ctx->iwdev->sc_dev.hmc_info->hmc_obj[I40E_HMC_IW_MR].
		      cnt / 8); l++) {
			m = k * 64 + l;
			status =
			    irdma_is_resource_allocated(dump_ctx->iwdev,
							dump_ctx->iwdev->
							allocated_mrs,
							m) ? 1 : 0;
			printk("%d ", status);
		}
		printk("\n");
	}

	printk(KERN_INFO "%s,%d === ARPs ===\n", __func__, __LINE__);

	for (k = 0; k < 8; k++) {
		for (l = 0;
		     l <
		     (dump_ctx->iwdev->sc_dev.hmc_info->
		      hmc_obj[I40E_HMC_IW_ARP].cnt / 8); l++) {
			m = k * 64 + l;
			status =
			    irdma_is_resource_allocated(dump_ctx->iwdev,
							dump_ctx->iwdev->
							allocated_arps,
							m) ? 1 : 0;
			printk("%d ", status);
		}
		printk("\n");
	}
}

/**
 * print_event_subscriptions_cnt -
 * @param event_ctx TODO:
 **/
static void print_event_subscriptions_cnt(struct icrdmadbg_evt_ctx *event_ctx)
{
	printk(KERN_INFO "%s,%d - %d %d %d %d %d %d\n", __func__, __LINE__,
	       event_ctx->evt_ref_cnt[IRDMADBG_EVENT_TYPE_AEQ],
	       event_ctx->evt_ref_cnt[IRDMADBG_EVENT_TYPE_AQ],
	       event_ctx->evt_ref_cnt[IRDMADBG_EVENT_TYPE_CEQ_ALL],
	       event_ctx->evt_ref_cnt[IRDMADBG_EVENT_TYPE_CEQ_ERR],
	       event_ctx->evt_ref_cnt[IRDMADBG_EVENT_TYPE_CQ0_ALL],
	       event_ctx->evt_ref_cnt[IRDMADBG_EVENT_TYPE_CQ0_ERR]);
}
#endif				/* IW_DEBUG_PRINT */
/*
 * Helper functions
 */
/*
 * return irdma_device with input ifindex
 * otherwise returns NULL
 */
static struct irdma_device *iwdbg_get_iwdev_by_index(int ifindex)
{
	struct icrdmadbg_device *tmp_iwdbg_device;
	struct irdma_device *tmp_iw_dev = NULL;

	spin_lock(&icrdmadbg_iw_dev_list_lock);
	list_for_each_entry(tmp_iwdbg_device, &icrdmadbg_iw_dev_list, node) {
		if (tmp_iwdbg_device->iwdev->netdev->ifindex == ifindex) {
			printk(KERN_INFO
			       "%s,%d IWARP DEVICE FOUND for ifindex %d\n",
			       __func__, __LINE__, ifindex);
			tmp_iw_dev = tmp_iwdbg_device->iwdev;
			break;
		}
	}

	spin_unlock(&icrdmadbg_iw_dev_list_lock);
	if (!tmp_iw_dev)
		printk(KERN_INFO
		       "%s,%d IWARP DEVICE NOT FOUND for ifindex %d\n",
		       __func__, __LINE__, ifindex);

	return tmp_iw_dev;
}

/*
 * Prepares dump ctx when first "dump it" is invoked
 */
static int iwdbg_prepare_dump_ctx(struct sk_buff *msg,
				  struct netlink_callback *cb,
				  struct icrdmadbg_hmc_dump_ctx *dump_ctx)
{
	int ret;
	struct nlattr *attrs[IRDMADBG_ATTR_COUNT];
	struct nlattr *attrs_hmc[IRDMADBG_ATTR_HMC_COUNT];

	dump_ctx->buffer_len = 0;

	nlmsg_parse(cb->nlh, GENL_HDRLEN, attrs, IRDMADBG_ATTR_MAX,
		    icrdmadbg_nl_policy);

	if (!attrs[IRDMADBG_ATTR_IFINDEX]) {
		printk(KERN_ERR "%s,%d attrs[IRDMADBG_ATTR_IFINDEX] \
			Problem !\n", __func__, __LINE__);
		goto err;
	}

	dump_ctx->ifindex = nla_get_u32(attrs[IRDMADBG_ATTR_IFINDEX]);

	if (!attrs[IRDMADBG_ATTR_HMC]) {
		printk(KERN_ERR "%s,%d attrs[IRDMADBG_ATTR_HMC] \
			Problem !\n", __func__, __LINE__);
		goto err;
	}

	ret = nla_parse_nested(attrs_hmc, IRDMADBG_ATTR_HMC_MAX,
			       attrs[IRDMADBG_ATTR_HMC],
			       icrdmadbg_nl_hmc_policy);
	if (ret != 0) {
		printk(KERN_ERR "%s,%d ret = nla_parse_nested for \
			attrs[IRDMADBG_ATTR_HMC] Problem,ret=%d !\n", __func__, __LINE__, ret);
		goto err;
	}

	if (!attrs_hmc[IRDMADBG_ATTR_HMC_OBJ_TYPE]) {
		printk(KERN_ERR "%s,%d attrs_hmc[IRDMADBG_ATTR_HMC_OBJ_TYPE] \
			Problem!\n", __func__, __LINE__);
		goto err;
	}

	dump_ctx->hmc_obj_type =
	    nla_get_u32(attrs_hmc[IRDMADBG_ATTR_HMC_OBJ_TYPE]);

	if (!attrs_hmc[IRDMADBG_ATTR_HMC_OBJ_ID_REQ])
		dump_ctx->hmc_obj_id_req = -1;
	else
		dump_ctx->hmc_obj_id_req =
		    nla_get_u32(attrs_hmc[IRDMADBG_ATTR_HMC_OBJ_ID_REQ]);

	if (!attrs_hmc[IRDMADBG_ATTR_HMC_VALID_ONLY])
		dump_ctx->valid_only = false;
	else
		dump_ctx->valid_only = true;

	if (!attrs_hmc[IRDMADBG_ATTR_HMC_QP_RAW])
		dump_ctx->qp_raw = false;
	else
		dump_ctx->qp_raw = true;

	if (!attrs_hmc[IRDMADBG_ATTR_HMC_QP_FREEZE])
		dump_ctx->qp_freeze = false;
	else
		dump_ctx->qp_freeze = true;

	if (!attrs_hmc[IRDMADBG_ATTR_HMC_QP_NO_CQP])
		dump_ctx->no_cqp = false;
	else
		dump_ctx->no_cqp = true;

	dump_ctx->iwdev = iwdbg_get_iwdev_by_index(dump_ctx->ifindex);

	if (!dump_ctx->iwdev)
		goto err;
#ifndef EXTERNAL_RELEASE
/*
 *      TODO:
 *      with ongoing driver/hw development
 *      more types of hmc object will be supported
 *      thus modification of below switch will be neccessary
 *
 *      unsigned long	 *allocated_qps;
 *         unsigned long	*allocated_cqs;
 *         unsigned long	*allocated_mrs;
 *         unsigned long	*allocated_arps;
 */
#endif
	switch (dump_ctx->hmc_obj_type) {
	case I40E_HMC_IW_QP:
		dump_ctx->resource_status_array =
		    dump_ctx->iwdev->allocated_qps;
		break;
	case I40E_HMC_IW_CQ:
		dump_ctx->resource_status_array =
		    dump_ctx->iwdev->allocated_cqs;
		break;
	case I40E_HMC_IW_MR:
		dump_ctx->resource_status_array =
		    dump_ctx->iwdev->allocated_mrs;
		break;
	case I40E_HMC_IW_ARP:
		dump_ctx->resource_status_array =
		    dump_ctx->iwdev->allocated_arps;
		break;
	case I40E_HMC_IW_APBVT_ENTRY:
		dump_ctx->resource_status_array = NULL;
		break;
	case I40E_HMC_IW_HTE:
		dump_ctx->resource_status_array = NULL;
		break;
	case I40E_HMC_IW_FSIMC:
		dump_ctx->resource_status_array = NULL;
		break;
	case I40E_HMC_IW_FSIAV:
		dump_ctx->resource_status_array = NULL;
		break;
	case I40E_HMC_IW_PBLE:
		dump_ctx->resource_status_array = NULL;
		break;
	default:
		dump_ctx->resource_status_array = NULL;

		printk(KERN_INFO
		       "%s,%d Resource status for hmc_obj_type=%d NOT supported\n",
		       __func__, __LINE__, dump_ctx->hmc_obj_type);
	}

	dump_ctx->hmc_obj_size =
	    dump_ctx->iwdev->sc_dev.hmc_info->hmc_obj[dump_ctx->hmc_obj_type].
	    size;

	dump_ctx->no_more_data_to_dump = false;

	if (dump_ctx->hmc_obj_id_req == -1) {
		dump_ctx->dump_cnt =
		    dump_ctx->iwdev->sc_dev.hmc_info->hmc_obj[dump_ctx->
							      hmc_obj_type].cnt;
		dump_ctx->dump_index = 0;
	} else {
		dump_ctx->dump_cnt = 1;
		dump_ctx->dump_index = dump_ctx->hmc_obj_id_req;
	}

	return IRDMA_SUCCESS;
 err:
	return -EINVAL;
}

/*
 * CSR access functions
 */
/*
 * Responisble for reading and wrting registers
 */
static int iwdbg_nl_cmd_reg(struct sk_buff *msg, struct genl_info *info)
{
	int ret;
	u32 ifindex;
	u32 regaddr;
	u32 regval_w, regval_r;
	u8 optype;
	struct irdma_hw *hw;
	struct nlattr *attrs_nested[IRDMADBG_ATTR_REG_COUNT];
	struct nlattr *attrs_response;
	struct irdma_device *iwdev;
	struct sk_buff *msg_resp;
	void *hdr;

	ret = -EINVAL;

	if (!info->attrs[IRDMADBG_ATTR_IFINDEX]) {
		printk(KERN_INFO
		       "%s,%d info->attrs[IRDMADBG_ATTR_IFINDEX] problem\n",
		       __func__, __LINE__);
		goto err;
	}

	if (!info->attrs[IRDMADBG_ATTR_REG]) {
		printk(KERN_INFO
		       "%s,%d info->attrs[IRDMADBG_ATTR_REG] problem\n",
		       __func__, __LINE__);
		goto err;
	}

	ret = nla_parse_nested(attrs_nested, IRDMADBG_ATTR_REG_MAX,
			       info->attrs[IRDMADBG_ATTR_REG],
			       icrdmadbg_nl_reg_policy);

	if (ret != 0) {
		printk(KERN_INFO "%s,%d ret = nla_parse_nested for \
			IRDMADBG_ATTR_REG_MAX:%d problem\n", __func__, __LINE__, ret);
		goto err;
	}

	if (!attrs_nested[IRDMADBG_ATTR_REG_OP_TYPE]) {
		printk(KERN_INFO
		       "%s,%d attrs_nested[IRDMADBG_ATTR_REG_OP_TYPE] problem\n",
		       __func__, __LINE__);
		goto err;
	}

	if (!attrs_nested[IRDMADBG_ATTR_REG_ADDR]) {
		printk(KERN_INFO
		       "%s,%d attrs_nested[IRDMADBG_ATTR_REG_ADDR] problem\n",
		       __func__, __LINE__);
		goto err;
	}

	if (attrs_nested[IRDMADBG_ATTR_REG_VALUE])
		regval_w = nla_get_u32(attrs_nested[IRDMADBG_ATTR_REG_VALUE]);

	optype = nla_get_u8(attrs_nested[IRDMADBG_ATTR_REG_OP_TYPE]);

	regaddr = nla_get_u32(attrs_nested[IRDMADBG_ATTR_REG_ADDR]);

	ifindex = nla_get_u32(info->attrs[IRDMADBG_ATTR_IFINDEX]);

	iwdev = iwdbg_get_iwdev_by_index(ifindex);
	if (!iwdev) {
		printk(KERN_ERR "%s,%d iwdev problem\n", __func__, __LINE__);
		goto err;
	}

	hw = iwdev->sc_dev.hw;
	if (!hw) {
		printk(KERN_ERR "%s,%d iwdev->sc_dev.hw problem\n",
		       __func__, __LINE__);
		goto err;
	}

	switch (optype) {
	case IRDMADBG_REG_OP_READ:
		break;
	case IRDMADBG_REG_OP_WRITE:
		wr32(hw, regaddr, regval_w);
		break;
	case IRDMADBG_REG_OP_SET_BITS:
		regval_r = rd32(hw, regaddr);
		wr32(hw, regaddr, regval_r | regval_w);
		break;
	case IRDMADBG_REG_OP_CLEAR_BITS:
		regval_r = rd32(hw, regaddr);
		wr32(hw, regaddr, regval_r & ~regval_w);
		break;
	default:
		printk(KERN_ERR "%s,%d operation\"%d\" not defined problem\n",
		       __func__, __LINE__, optype);
		goto err;
	}

	/* prepare response message for user space for each operation type */

	regval_r = rd32(hw, regaddr);

	msg_resp = nlmsg_new(NLMSG_GOODSIZE, GFP_KERNEL);
	if (msg_resp == NULL) {
		printk(KERN_ERR "%s,%d msg_resp problem\n", __func__, __LINE__);
		ret = -ENOMEM;
		goto err;
	}

	hdr = genlmsg_put(msg_resp, info->snd_pid, info->snd_seq,
			  &icrdmadbg_nl_family, NLM_F_ACK, IRDMADBG_CMD_REG);
	if (!hdr) {
		printk(KERN_ERR "%s,%d hdr genlmsg_put problem\n",
		       __func__, __LINE__);

		ret = -EMSGSIZE;
		goto err_put;
	}

	ret = nla_put_u32(msg_resp, IRDMADBG_ATTR_IFINDEX, ifindex);
	if (ret)
		goto err_put;

	attrs_response = nla_nest_start(msg_resp, IRDMADBG_ATTR_REG);
	if (!attrs_response) {
		printk(KERN_ERR "%s,%d attrs_response problem\n",
		       __func__, __LINE__);
		return -EMSGSIZE;
	}

	ret = nla_put_u8(msg_resp, IRDMADBG_ATTR_REG_OP_TYPE, optype);
	if (ret)
		goto err_put;

	ret = nla_put_u32(msg_resp, IRDMADBG_ATTR_REG_ADDR, regaddr);
	if (ret)
		goto err_put;

	ret = nla_put_u32(msg_resp, IRDMADBG_ATTR_REG_VALUE, regval_r);
	if (ret)
		goto err_put;

	nla_nest_end(msg_resp, attrs_response);
	genlmsg_end(msg_resp, hdr);

	ret = genlmsg_unicast(genl_info_net(info), msg_resp, info->snd_pid);
	if (ret < 0)
		printk(KERN_ERR "%s,%d ret=%d sending back to user problem\n",
		       __func__, __LINE__, ret);

	return ret;

 err_put:
	printk(KERN_ERR "%s,%d nla_put problem\n", __func__, __LINE__);
	nlmsg_free(msg_resp);

 err:
	return ret;
}

/*
 * Functions for dumping HMC objects
 */
/*
 * uploads single QP ctx to dest. buffer
 * in either RAW or structured format
 */
static int iwdbg_cqp_upload_ctx(struct icrdmadbg_hmc_dump_ctx *dump_ctx,
				void *buffer)
{
	struct irdma_sc_dev *dev;
	u64 scratch = 0;
	struct irdma_upload_context_info info;
	struct irdma_dma_mem dma_mem;
	struct irdma_ccq_cqe_info iwdbg_compl_info;
	int ret = IRDMA_SUCCESS;
	unsigned long flags;

	dev = &dump_ctx->iwdev->sc_dev;
	if (!dev) {
		DEBUGOUT("dump_ctx->iwdev->sc_dev problem\n");
		ret = IRDMA_ERR_BAD_PTR;
		goto out;
	}

	if (!buffer) {
		DEBUGOUT("iwdbg_cqp_upload_ctx buffer problem\n");
		ret = IRDMA_ERR_BAD_PTR;
		goto out;
	}

	ret = irdma_allocate_dma_mem(dev->hw, &dma_mem,
				     i40e_mem_reserved,
				     (u64) dump_ctx->hmc_obj_size, PAGE_SIZE);
	if (ret != IRDMA_SUCCESS) {
		DEBUGOUT("irdma_allocate_dma_mem: problem\n");
		goto out;
	}

	info.freeze_qp = dump_ctx->qp_freeze;
	info.raw_format = dump_ctx->qp_raw;
	info.buf_pa = dma_mem.pa;
/*
 *	IRDMA_QP_TYPE_IWARP
 *	IRDMA_QP_TYPE_UDA
 */
	info.qp_type = IRDMA_QP_TYPE_UDA;
	info.qp_id = dump_ctx->dump_index;

	spin_lock_irqsave(&dump_ctx->iwdev->cqp.lock, flags);

	ret =
	    dev->iw_priv_qp_ops->iwarp_qp_upload_context(dev, &info, scratch,
							 true);
	if (ret != IRDMA_SUCCESS) {
		DEBUGOUT("iwarp_qp_upload_context problem\n");
		spin_unlock_irqrestore(&dump_ctx->iwdev->cqp.lock, flags);
		goto out;
	}

	if (dump_ctx->iwdev->event_ctx->evt_ref_cnt[IRDMADBG_EVENT_TYPE_CQ0_ALL]
	    || dump_ctx->iwdev->event_ctx->
	    evt_ref_cnt[IRDMADBG_EVENT_TYPE_CQ0_ERR]) {
		ret = irdma_sc_poll_for_cqp_op_done(dev->cqp, IRDMA_CQP_OP_UPLOAD_CONTEXT,
						    &iwdbg_compl_info);

		spin_unlock_irqrestore(&dump_ctx->iwdev->cqp.lock, flags);

		if ((ret == IRDMA_SUCCESS || ret == IRDMA_ERR_CQP_COMPL_ERROR)
		    && (dump_ctx->iwdev->event_ctx->
			evt_ref_cnt[IRDMADBG_EVENT_TYPE_CQ0_ALL]
			|| (dump_ctx->iwdev->event_ctx->
			    evt_ref_cnt[IRDMADBG_EVENT_TYPE_CQ0_ERR]
			    && (ret == IRDMA_ERR_CQP_COMPL_ERROR))))
			icrdmadbg_process_cqp_event(dump_ctx->iwdev,
						   &iwdbg_compl_info);
	} else {
		ret = irdma_sc_poll_for_cqp_op_done(dev->cqp, IRDMA_CQP_OP_UPLOAD_CONTEXT,
						    NULL);

		spin_unlock_irqrestore(&dump_ctx->iwdev->cqp.lock, flags);
	}
	if (ret != IRDMA_SUCCESS) {
		DEBUGOUT("poll_for_cqp_op_done: problem\n");
		goto out;
	}

	memcpy(buffer, dma_mem.va, dump_ctx->hmc_obj_size);

 out:
	irdma_free_dma_mem(dev->hw, &dma_mem);

	return ret;
}

/*
 * Dumps IRDMADBG_HMC_OBJ_VALID and IRDMADBG_HMC_OBJ_NOT_VALID objects
 */
static int iwdbg_nl_driver_mode_dump_all(struct icrdmadbg_hmc_dump_ctx *dump_ctx,
					 struct sk_buff *msg,
					 struct netlink_callback *cb)
{
	int ret, i;
	struct nlattr *attrs_hmc_out;
	struct i40e_hmc_iw_dump_obj_info hmc_dump_info;
	void *resp_msg_hdr = NULL;
	u8 *ptr;

	resp_msg_hdr =
	    genlmsg_put(msg, NETLINK_CB(cb->skb).pid, cb->nlh->nlmsg_seq,
			&icrdmadbg_nl_family, NLM_F_MULTI, IRDMADBG_CMD_HMC);
	if (!resp_msg_hdr) {
		printk(KERN_ERR "%s,%d ret = resp_msg_hdr Problem !\n",
		       __func__, __LINE__);
		goto err;
	}

	/* calculate this only once */
	if (!dump_ctx->max_obj)
		dump_ctx->max_obj = (skb_tailroom(msg) - IRDMADBG_MESSAGE_RESERVE_BYTES) / (dump_ctx->hmc_obj_size + sizeof(u8));	/* +1byte for uint8 of status */

	if (dump_ctx->dump_cnt <= dump_ctx->max_obj) {
		hmc_dump_info.count = dump_ctx->dump_cnt;
		dump_ctx->no_more_data_to_dump = true;
	} else {
		hmc_dump_info.count = dump_ctx->max_obj;
	}

	if (hmc_dump_info.count * (dump_ctx->hmc_obj_size + sizeof(u8)) != dump_ctx->buffer_len) {	/* +1byte for uint8 of status */
		if (dump_ctx->buffer_len != 0) {
			IW_KFREE(dump_ctx->buffer);
			dump_ctx->buffer = NULL;
		}

		dump_ctx->buffer_len = (dump_ctx->hmc_obj_size + sizeof(u8)) * hmc_dump_info.count;	/*+1byte for uint8 of status */
		dump_ctx->buffer = kzalloc(dump_ctx->buffer_len, GFP_KERNEL);

		if (!dump_ctx->buffer) {
			printk(KERN_ERR "%s,%d dump_ctx->buffer problem\n",
			       __func__, __LINE__);
			ret = -ENOMEM;
			goto err;
		}
	}

	hmc_dump_info.start_idx = dump_ctx->dump_index;
	hmc_dump_info.hmc_info = dump_ctx->iwdev->sc_dev.hmc_info;
	hmc_dump_info.rsrc_type = dump_ctx->hmc_obj_type;
	hmc_dump_info.dest_buf_size = dump_ctx->buffer_len - (hmc_dump_info.count * sizeof(u8));	/* - hmc_dump_info.count, (get rid of status bytes for each object) */
	hmc_dump_info.dest_buf = (u8 *) dump_ctx->buffer + (hmc_dump_info.count * sizeof(u8));	/*+ hmc_dump_info.count, add offset for status bytes */

	ret =
	    dump_ctx->iwdev->sc_dev.hmc_ops->dump_hmc_obj(dump_ctx->iwdev->
							  sc_dev.hw,
							  &hmc_dump_info);

	if (ret == IRDMA_ERR_BACKING_PAGE_ERROR)
		printk(KERN_ERR
		       "%s,%d IRDMA_ERR_BACKING_PAGE_ERROR hmc_ops->dump_hmc_obj:%d\n",
		       __func__, __LINE__, ret);
	if (ret != IRDMA_ERR_BACKING_PAGE_ERROR && ret != IRDMA_SUCCESS) {
		printk(KERN_ERR "%s,%d hmc_ops->dump_hmc_obj:%d\n", __func__,
		       __LINE__, ret);
		goto err;
	}

	dump_ctx->dump_cnt -= hmc_dump_info.count;
	dump_ctx->dump_index += hmc_dump_info.count;

	ptr = dump_ctx->buffer;
	for (i = 0; i < hmc_dump_info.count; i++) {
		if (dump_ctx->resource_status_array) {	/* status for tracked HMC object only */
			if (irdma_is_resource_allocated(dump_ctx->iwdev,
							dump_ctx->
							resource_status_array,
							i +
							hmc_dump_info.
							start_idx))

				ptr[i] = (u8) IRDMADBG_HMC_OBJ_VALID;
			else

				ptr[i] = (u8) IRDMADBG_HMC_OBJ_NOT_VALID;
		} else {	/* status for not tracked HMC object only */
			ptr[i] = (u8) IRDMADBG_HMC_OBJ_STATUS_UNKNOWN;
		}
	}

	attrs_hmc_out = nla_nest_start(msg, IRDMADBG_ATTR_HMC);
	ret =
	    nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_SIZE,
			dump_ctx->hmc_obj_size);
	if (ret)
		goto err_put;

	ret =
	    nla_put(msg, IRDMADBG_ATTR_HMC_OBJ_DATA, dump_ctx->buffer_len,
		    dump_ctx->buffer);
	if (ret)
		goto err_put;

	ret =
	    nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_START_ID,
			hmc_dump_info.start_idx);
	if (ret)
		goto err_put;
	ret = nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_CNT, hmc_dump_info.count);
	if (ret)
		goto err_put;

	nla_nest_end(msg, attrs_hmc_out);
	ret = genlmsg_end(msg, resp_msg_hdr);
	if (ret < 0) {
		printk(KERN_ERR "%s,%d iwdbg_nl_driver_mode_dump_all ret=%d\n",
		       __func__, __LINE__, ret);

		goto err;
	}

	return msg->len;

 err_put:
	printk(KERN_ERR "%s,%d nla_put binary data ret=%d\n",
	       __func__, __LINE__, ret);

 err:
	/* we can be here already having a buffer */
	if (dump_ctx->buffer) {
		IW_KFREE(dump_ctx->buffer);
		dump_ctx->buffer = NULL;
	}

	return ret;
}

/*
 * Dumps IRDMADBG_HMC_OBJ_VALID only objects
 */
static int iwdbg_nl_driver_mode_dump_valid_only(struct icrdmadbg_hmc_dump_ctx
						*dump_ctx, struct sk_buff *msg,
						struct netlink_callback *cb)
{
	int ret, packed_obj;
	struct nlattr *attrs_hmc_out;
	struct i40e_hmc_iw_dump_obj_info hmc_dump_info;
	void *resp_msg_hdr = NULL;
	u8 *ptr;

	/* dump only valid objects of given type, placed in buffer in following manner:
	 * |_4B of index_|____data with len =hmc_obj_size_____.....
	 * |_4B of index_|____data with len =hmc_obj_size_____|
	 */
	resp_msg_hdr =
	    genlmsg_put(msg, NETLINK_CB(cb->skb).pid, cb->nlh->nlmsg_seq,
			&icrdmadbg_nl_family, NLM_F_MULTI, IRDMADBG_CMD_HMC);
	if (!resp_msg_hdr) {
		printk(KERN_ERR "%s,%d ret = resp_msg_hdr \
			Problem !\n", __func__, __LINE__);

		ret = -EMSGSIZE;
		goto err;
	}

	if (!dump_ctx->max_obj) {
		dump_ctx->max_obj = (skb_tailroom(msg) - IRDMADBG_MESSAGE_RESERVE_BYTES) / (dump_ctx->hmc_obj_size + sizeof(u32));	/* +4bytes for u32 of id */
	}

	if ((dump_ctx->buffer_len == 0) && !dump_ctx->buffer) {
		/*+4bytes for u32 of id */
		dump_ctx->buffer_len =
		    (dump_ctx->hmc_obj_size + sizeof(u32)) * dump_ctx->max_obj;
		dump_ctx->buffer = kzalloc(dump_ctx->buffer_len, GFP_KERNEL);
		if (!dump_ctx->buffer) {
			printk(KERN_ERR "%s,%d dump_ctx->buffer problem\n",
			       __func__, __LINE__);

			ret = -ENOMEM;
			goto err;
		}
	}
	packed_obj = 0;
	ptr = dump_ctx->buffer;

	hmc_dump_info.count = 1;
	hmc_dump_info.hmc_info = dump_ctx->iwdev->sc_dev.hmc_info;
	hmc_dump_info.rsrc_type = dump_ctx->hmc_obj_type;
	hmc_dump_info.dest_buf_size = dump_ctx->hmc_obj_size;

	if (dump_ctx->resource_status_array) {
		while (dump_ctx->dump_index < dump_ctx->dump_cnt) {
			if (irdma_is_resource_allocated(dump_ctx->iwdev,
							dump_ctx->
							resource_status_array,
							dump_ctx->dump_index)) {
				((u32 *) ptr)[0] = dump_ctx->dump_index;	/* set index in destination buffer */
				ptr += sizeof(u32);	/* u8 ptr + size of u32 calculate size for object data starting point */

				hmc_dump_info.start_idx = dump_ctx->dump_index;
				hmc_dump_info.dest_buf = ptr;
				ret =
				    dump_ctx->iwdev->sc_dev.hmc_ops->
				    dump_hmc_obj(dump_ctx->iwdev->sc_dev.hw,
						 &hmc_dump_info);
				if (ret != IRDMA_SUCCESS) {
					printk(KERN_INFO
					       "%s,%d hmc_ops->dump_hmc_obj:%d\n",
					       __func__, __LINE__, ret);

					goto err;
				}

				ptr += dump_ctx->hmc_obj_size;
				packed_obj++;
			}

			dump_ctx->dump_index++;
			if (dump_ctx->dump_index == dump_ctx->dump_cnt)
				dump_ctx->no_more_data_to_dump = true;

			if (packed_obj == dump_ctx->max_obj)
				break;
		}
	} else {
		/* there is no tracking of valid objects for given type */
		dump_ctx->no_more_data_to_dump = true;
		printk(KERN_INFO "%s,%d No status for HMC object type=%d\n",
		       __func__, __LINE__, dump_ctx->hmc_obj_type);
	}

	if (packed_obj > 0) {
		attrs_hmc_out = nla_nest_start(msg, IRDMADBG_ATTR_HMC);

		ret =
		    nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_SIZE,
				dump_ctx->hmc_obj_size);
		if (ret)
			goto err_put;

		/* set lengt of buffer to real length of dumped (obj_size + 4bytes of index)* dumped obj cnt */
		ret = nla_put(msg, IRDMADBG_ATTR_HMC_OBJ_DATA, packed_obj *
			      (dump_ctx->hmc_obj_size + sizeof(u32)),
			      dump_ctx->buffer);

		ret = nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_CNT, packed_obj);
		if (ret)
			goto err_put;

		ret = nla_put_flag(msg, IRDMADBG_ATTR_HMC_VALID_ONLY);
		if (ret)
			goto err_put;

		nla_nest_end(msg, attrs_hmc_out);

		genlmsg_end(msg, resp_msg_hdr);

		return msg->len;
	}

	/* if we are  here it means nothing was packed and flag
	 * "no_more_data_to_dump" is set, return 1 will cause call to
	 * iwdbg_nl_cmd_hmc_dump one last time and flag
	 *"no_more_data_to_dump" will free dump_ctx resources
	 */
	return 1;

 err_put:
	printk(KERN_ERR "%s,%d nla_put binary data ret=%d\n",
	       __func__, __LINE__, ret);

 err:
	/* we can be here already having buffer */
	if (dump_ctx->buffer)
		IW_KFREE(dump_ctx->buffer);

	IW_KFREE(dump_ctx);

	return ret;
}

/*
 * Dumps IRDMADBG_HMC_OBJ_VALID and IRDMADBG_HMC_OBJ_NOT_VALID objects
 */
static int iwdbg_nl_cqp_mode_dump_qp_valid_only(struct icrdmadbg_hmc_dump_ctx
						*dump_ctx, bool raw_format,
						struct sk_buff *msg,
						struct netlink_callback *cb)
{
	int ret, packed_obj;
	struct nlattr *attrs_hmc_out;
	void *resp_msg_hdr = NULL;
	u8 *ptr;

	/* dump only valid objects of given type, placed in buffer in following manner:
	 * |_4B of index_|____data with len =hmc_obj_size_____.....
	 * |_4B of index_|____data with len =hmc_obj_size_____|
	 */
	resp_msg_hdr =
	    genlmsg_put(msg, NETLINK_CB(cb->skb).pid, cb->nlh->nlmsg_seq,
			&icrdmadbg_nl_family, NLM_F_MULTI, IRDMADBG_CMD_HMC);
	if (!resp_msg_hdr) {
		printk(KERN_ERR "%s,%d ret = resp_msg_hdr Problem !\n",
		       __func__, __LINE__);

		ret = -EMSGSIZE;
		goto err;
	}

	/* default is RAW foramat for QP ctx -- 512B */
	dump_ctx->hmc_obj_size =
	    dump_ctx->iwdev->sc_dev.hmc_info->hmc_obj[dump_ctx->hmc_obj_type].
	    size;

	/* check format type, 128 B is for structured format */
	if (!raw_format)
		dump_ctx->hmc_obj_size = IRDMADBG_QP_CTX_STRUCTERED_SIZE;

	/* calculate this only once */
	if (!dump_ctx->max_obj) {
		/* +4bytes for u32 of id */
		dump_ctx->max_obj =
		    (skb_tailroom(msg) -
		     IRDMADBG_MESSAGE_RESERVE_BYTES) / (dump_ctx->hmc_obj_size +
							sizeof(u32));
	}

	/* prepare buffer only once,use maximu size we can pack in one message */
	if ((dump_ctx->buffer_len == 0) && !dump_ctx->buffer) {
		/* +4bytes for u32 of id */
		dump_ctx->buffer_len = (dump_ctx->hmc_obj_size + sizeof(u32)) *
		    dump_ctx->max_obj;
		dump_ctx->buffer = kzalloc(dump_ctx->buffer_len, GFP_KERNEL);
		if (!dump_ctx->buffer) {
			printk(KERN_ERR "%s,%d dump_ctx->buffer problem\n",
			       __func__, __LINE__);

			ret = -ENOMEM;
			goto err;
		}
	}
	packed_obj = 0;
	ptr = dump_ctx->buffer;

	if (dump_ctx->dump_index == 0) {
		printk(KERN_INFO
		       "%s,%d Skipping cqp_upload operation for QP id=%d\n",
		       __func__, __LINE__, dump_ctx->dump_index);
		dump_ctx->dump_index++;
	}

	while (dump_ctx->dump_index < dump_ctx->dump_cnt) {
		if (irdma_is_resource_allocated(dump_ctx->iwdev,
						dump_ctx->resource_status_array,
						dump_ctx->dump_index)) {
			((u32 *) ptr)[0] = dump_ctx->dump_index;	/* set index in destination buffer */
			ptr += sizeof(u32);	/* u8 *ptr + size of u32 calculate size for object data starting point */

			ret = iwdbg_cqp_upload_ctx(dump_ctx, ptr);
			if (ret != IRDMA_SUCCESS) {
				printk(KERN_INFO
				       "%s,%d iwdbg_cqp_upload_ctx ret=%d\n",
				       __func__, __LINE__, ret);
				goto err;
			}

			ptr += dump_ctx->hmc_obj_size;
			packed_obj++;
		}

		dump_ctx->dump_index++;
		if (dump_ctx->dump_index == dump_ctx->dump_cnt)
			dump_ctx->no_more_data_to_dump = true;

		if (packed_obj == dump_ctx->max_obj)
			break;
	}

	if (packed_obj > 0) {
		attrs_hmc_out = nla_nest_start(msg, IRDMADBG_ATTR_HMC);
		ret =
		    nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_SIZE,
				dump_ctx->hmc_obj_size);
		if (ret)
			goto err_put;

		/* set lengt of buffer to real lengt of dumped (obj_size + 4bytes of index)* dumped obj cnt */
		ret =
		    nla_put(msg, IRDMADBG_ATTR_HMC_OBJ_DATA,
			    packed_obj * (dump_ctx->hmc_obj_size + sizeof(u32)),
			    dump_ctx->buffer);
		if (ret)
			goto err_put;

		ret = nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_CNT, packed_obj);
		if (ret)
			goto err_put;

		ret = nla_put_flag(msg, IRDMADBG_ATTR_HMC_VALID_ONLY);
		if (ret)
			goto err_put;

		nla_nest_end(msg, attrs_hmc_out);

		ret = genlmsg_end(msg, resp_msg_hdr);

		return msg->len;
	}

	/* if we are  here it means nothing was packed and flag
	 * "no_more_data_to_dump" is set, return 1 will cause call to
	 * iwdbg_nl_cmd_hmc_dump one last time and flag
	 *"no_more_data_to_dump" will free dump_ctx resources
	 */
	return 1;

 err_put:
	printk(KERN_ERR "%s,%d nla_put binary data ret=%d\n",
	       __func__, __LINE__, ret);

 err:
	/* we can be here already having buffer */
	if (dump_ctx->buffer)
		IW_KFREE(dump_ctx->buffer);

	IW_KFREE(dump_ctx);
	return ret;
}

/*
 * Dumps single QP context via CQP ulpoad operaion
 */
static int iwdbg_nl_cqp_mode_dump_single_qp_ctx(struct icrdmadbg_hmc_dump_ctx
						*dump_ctx, bool raw_format,
						struct sk_buff *msg,
						struct netlink_callback *cb)
{
	int ret;
	struct nlattr *attrs_hmc_out;
	void *resp_msg_hdr = NULL;

	resp_msg_hdr =
	    genlmsg_put(msg, NETLINK_CB(cb->skb).pid, cb->nlh->nlmsg_seq,
			&icrdmadbg_nl_family, NLM_F_MULTI, IRDMADBG_CMD_HMC);
	if (!resp_msg_hdr) {
		printk(KERN_ERR "%s,%d ret = resp_msg_hdr Problem !\n",
		       __func__, __LINE__);

		ret = -EMSGSIZE;
		goto err;
	}

	/* default is RAW foramat for QP ctx -- 512B */
	dump_ctx->hmc_obj_size =
	    dump_ctx->iwdev->sc_dev.hmc_info->hmc_obj[dump_ctx->hmc_obj_type].
	    size;

	/* check format type, 128 B is for structured format */
	if (!raw_format)
		dump_ctx->hmc_obj_size = IRDMADBG_QP_CTX_STRUCTERED_SIZE;

	dump_ctx->buffer_len = dump_ctx->hmc_obj_size;

	dump_ctx->buffer = kzalloc(dump_ctx->buffer_len, GFP_KERNEL);
	if (!dump_ctx->buffer) {
		printk(KERN_ERR "%s,%d dump_ctx->buffer problem\n",
		       __func__, __LINE__);
		ret = -ENOMEM;
		goto err;
	}

	ret = iwdbg_cqp_upload_ctx(dump_ctx, dump_ctx->buffer);
	if (ret != IRDMA_SUCCESS) {
		printk(KERN_ERR "%s,%d iwdbg_cqp_upload_ctx problem\n",
		       __func__, __LINE__);
		goto err;
	}

	attrs_hmc_out = nla_nest_start(msg, IRDMADBG_ATTR_HMC);

	ret =
	    nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_SIZE,
			dump_ctx->hmc_obj_size);
	if (ret)
		goto err_put;

	ret = nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_CNT, 1);
	if (ret)
		goto err_put;

	ret =
	    nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_START_ID,
			dump_ctx->dump_index);
	if (ret)
		goto err_put;

	ret =
	    nla_put(msg, IRDMADBG_ATTR_HMC_OBJ_DATA, dump_ctx->buffer_len,
		    dump_ctx->buffer);
	if (ret)
		goto err_put;

	nla_nest_end(msg, attrs_hmc_out);

	ret = genlmsg_end(msg, resp_msg_hdr);

	dump_ctx->no_more_data_to_dump = true;

	return msg->len;

 err_put:
	printk(KERN_ERR "%s,%d nla_put binary data ret=%d\n",
	       __func__, __LINE__, ret);
 err:
	/* we can be here already having buffer */
	if (dump_ctx->buffer)
		IW_KFREE(dump_ctx->buffer);

	IW_KFREE(dump_ctx);
	return ret;
}

/*
 * Dumps APBVTE object
 * send APBVTE in 4 parts to user space debug app
 */
static int iwdbg_nl_driver_mode_dump_apbvt(struct icrdmadbg_hmc_dump_ctx
					   *dump_ctx, struct sk_buff *msg,
					   struct netlink_callback *cb)
{
	struct i40e_hmc_iw_dump_obj_info hmc_dump_info;
	struct nlattr *attrs_hmc_out;
	void *resp_msg_hdr = NULL;
	int ret;

	/*
	 *      prepare context buffer for APVBT dump
	 */
	if (dump_ctx->buffer_len == 0) {
		dump_ctx->buffer_len =
		    dump_ctx->iwdev->sc_dev.hmc_info->
		    hmc_obj[I40E_HMC_IW_APBVT_ENTRY].size;

		dump_ctx->buffer = kzalloc(dump_ctx->buffer_len, GFP_KERNEL);
		if (!dump_ctx->buffer) {
			printk(KERN_ERR "%s,%d dump_ctx->buffer problem\n",
			       __func__, __LINE__);
			goto err;
		}

		hmc_dump_info.count = 1;
		hmc_dump_info.hmc_info = dump_ctx->iwdev->sc_dev.hmc_info;
		hmc_dump_info.rsrc_type = I40E_HMC_IW_APBVT_ENTRY;
		hmc_dump_info.dest_buf_size = dump_ctx->buffer_len;
		hmc_dump_info.start_idx = 0;
		hmc_dump_info.dest_buf = dump_ctx->buffer;

		ret =
		    dump_ctx->iwdev->sc_dev.hmc_ops->dump_hmc_obj(dump_ctx->
								  iwdev->sc_dev.
								  hw,
								  &hmc_dump_info);
		if (ret != IRDMA_SUCCESS) {
			printk(KERN_INFO "%s,%d hmc_ops->dump_hmc_obj:%d\n",
			       __func__, __LINE__, ret);
			IW_KFREE(dump_ctx->buffer);
			IW_KFREE(dump_ctx);
			return ret;
		}
		/*
		 *      send APBVTE in 4 messaages
		 *      2048B of APBVTE each
		 */
		dump_ctx->dump_cnt =
		    dump_ctx->buffer_len / IRDMADBG_APBVT_PART_SIZE;
		dump_ctx->dump_index = 0;
	}

	resp_msg_hdr = genlmsg_put(msg,
				   NETLINK_CB(cb->skb).pid, cb->nlh->nlmsg_seq,
				   &icrdmadbg_nl_family, NLM_F_MULTI,
				   IRDMADBG_CMD_HMC);
	if (!resp_msg_hdr) {
		printk(KERN_ERR "%s,%d resp_msg_hdr Problem !\n",
		       __func__, __LINE__);
		IW_KFREE(dump_ctx->buffer);
		goto err;
	}

	attrs_hmc_out = nla_nest_start(msg, IRDMADBG_ATTR_HMC);

	ret =
	    nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_SIZE,
			IRDMADBG_APBVT_PART_SIZE);
	if (ret)
		goto err_put;

	ret =
	    nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_START_ID,
			dump_ctx->dump_index);
	if (ret)
		goto err_put;

	ret = nla_put_u32(msg, IRDMADBG_ATTR_HMC_OBJ_CNT, 1);
	if (ret)
		goto err_put;

	ret = nla_put(msg, IRDMADBG_ATTR_HMC_OBJ_DATA,
		      IRDMADBG_APBVT_PART_SIZE, dump_ctx->buffer +
		      (dump_ctx->dump_index * IRDMADBG_APBVT_PART_SIZE *
		       sizeof(u8)));
	if (ret)
		goto err_put;

	nla_nest_end(msg, attrs_hmc_out);

	dump_ctx->dump_index++;
	if (dump_ctx->dump_index == IRDMADBG_APBVT_PART_CNT)
		dump_ctx->no_more_data_to_dump = true;

	genlmsg_end(msg, resp_msg_hdr);

	return msg->len;

 err_put:
	printk(KERN_ERR "%s,%d nla_put binary data ret=%d\n",
	       __func__, __LINE__, ret);
	IW_KFREE(dump_ctx->buffer);

 err:
	IW_KFREE(dump_ctx);

	return -EINVAL;
}

/*
 * "dump it" callback function
 */
static int iwdbg_nl_cmd_hmc_dump(struct sk_buff *msg,
				 struct netlink_callback *cb)
{
	int ret;
	struct icrdmadbg_hmc_dump_ctx *dump_ctx;

	/* cb->args[0] is used to keep addres of context info between
	 *  calls to iwdbg_nl_cmd_hmc_dump which is
	 * being called as long as it returns ret(in fact msg->len) > 0 (meaning data was sent out)
	 * first dump request prepares dump context to store info about request
	 * next calls are using and modifing  this context to complete dump request
	 */
	if (cb->args[0] == NULL) {
		dump_ctx = kzalloc(sizeof(*dump_ctx), GFP_KERNEL);
		if (!dump_ctx) {
			printk(KERN_ERR "%s,%d dump_ctx kzalloc Problem !\n",
			       __func__, __LINE__);
			goto err;
		}

		ret = iwdbg_prepare_dump_ctx(msg, cb, dump_ctx);
		if (ret != IRDMA_SUCCESS)
			goto err_dump_ctx;
#ifdef IW_DEBUG_PRINT

		print_resources_status(dump_ctx);
#endif				/* IW_DEBUG_PRINT */

		cb->args[0] = dump_ctx;
	} else {
		dump_ctx = (struct icrdmadbg_hmc_dump_ctx *)cb->args[0];
		if (dump_ctx->no_more_data_to_dump) {
			printk(KERN_INFO "%s,%d === Dump is done\n", __func__,
			       __LINE__);

			if (dump_ctx->buffer)
				IW_KFREE(dump_ctx->buffer);
			IW_KFREE(dump_ctx);

			return msg->len;
		}
	}

	/* treat APBVTE dump in special way */
	if (dump_ctx->hmc_obj_type == I40E_HMC_IW_APBVT_ENTRY) {
		ret = iwdbg_nl_driver_mode_dump_apbvt(dump_ctx, msg, cb);
		return ret;

		/* treat QP ctx dump in special way */
	} else if (dump_ctx->hmc_obj_type == I40E_HMC_IW_QP) {
		/* do not use cqp upload for QP ctx dump */
		if (dump_ctx->no_cqp) {
			if (dump_ctx->valid_only) {
				ret =
				    iwdbg_nl_driver_mode_dump_valid_only
				    (dump_ctx, msg, cb);
				return ret;
			} else {
				ret =
				    iwdbg_nl_driver_mode_dump_all(dump_ctx, msg,
								  cb);
				return ret;
			}
			/* use cqp upload method */
		} else {
			if (dump_ctx->valid_only) {
				ret =
				    iwdbg_nl_cqp_mode_dump_qp_valid_only
				    (dump_ctx, dump_ctx->qp_raw, msg, cb);
				return ret;
			} else {
				ret =
				    iwdbg_nl_cqp_mode_dump_single_qp_ctx
				    (dump_ctx, dump_ctx->qp_raw, msg, cb);
				return ret;
			}
		}
		/* Other objects than APBVTE or QP ctx */
	} else {
		if (dump_ctx->valid_only) {
			ret = iwdbg_nl_driver_mode_dump_valid_only(dump_ctx,
								   msg, cb);
			return ret;
		} else {
			ret = iwdbg_nl_driver_mode_dump_all(dump_ctx, msg, cb);
			return ret;
		}
	}
 err_dump_ctx:
	IW_KFREE(dump_ctx);
 err:
	return ret;
}

/*
 * Functions for event processing
 */
static int iwdbg_add_monitor_to_wait_evt_ctx(struct icrdmadbg_evt_ctx
					     *wait_evt_ctx, int ifindex,
					     struct nlattr *attrs_nested,
					     u32 pid)
{
	u32 waiting_ctx_cnt = 0;
	u32 evt_type;
	struct icrdmadbg_evt_ctx *tmp;
	bool wait_ctx_found = false;
	struct nlattr *tmp_attr_event_type;
	int rem;

	list_for_each_entry(tmp, &icrdmadbg_evt_ctx_wait_list, node) {
		/* go through whole list
		 * to get number of all waiting contexts
		 */
		waiting_ctx_cnt++;

		if (tmp->ifindex == ifindex) {
			wait_ctx_found = true;
			/* dont need wait_evt_ctx any more */
			IW_KFREE(wait_evt_ctx);
			wait_evt_ctx = tmp;

			printk(KERN_INFO
			       "%s,%d WAIT CONTEXT found for ifindex %d\n",
			       __func__, __LINE__, ifindex);
		}
	}

	if (!wait_ctx_found) {
		if (waiting_ctx_cnt == IRDMADBG_EVENT_CFG_WAITLIST_MAX_SIZE) {
			printk(KERN_ERR
			       "%s:%d IRDMADBG_EVENT_CFG_WAITLIST_MAX_SIZE=%d reached!\n",
			       __func__, __LINE__,
			       IRDMADBG_EVENT_CFG_WAITLIST_MAX_SIZE);

			return IRDMA_ERR_CONFIG;
		}

		printk(KERN_INFO "%s,%d CREATING WAIT CONTEXT for ifindex %d\n",
		       __func__, __LINE__, ifindex);

		wait_evt_ctx->ifindex = ifindex;
		list_add(&(wait_evt_ctx->node), &icrdmadbg_evt_ctx_wait_list);
	}

	printk(KERN_INFO
	       "%s,%d UPDATING WAIT CONTEXT for ifindex %d from PID %d\n",
	       __func__, __LINE__, ifindex, pid);

	wait_evt_ctx->client_ref_cnt++;

	printk(KERN_INFO "%s,%d %d client(s) waiting for ifindex %d\n",
	       __func__, __LINE__, wait_evt_ctx->client_ref_cnt, ifindex);

	nla_for_each_nested(tmp_attr_event_type, attrs_nested, rem) {
		evt_type = nla_get_u32(tmp_attr_event_type);
		if (evt_type < IRDMADBG_EVENT_TYPE_COUNT)
			wait_evt_ctx->evt_ref_cnt[evt_type] += 1;
	}

	return IRDMA_SUCCESS;
}

/**
 * iwdbg_del_monitor_from_wait_evt_ctx -
 * @param ifindex TODO
 * @param attrs_nested TODO
 * @param pid TODO
 * @return TODO:
 **/
int iwdbg_del_monitor_from_wait_evt_ctx(int ifindex,
					struct nlattr *attrs_nested, u32 pid)
{
	struct icrdmadbg_evt_ctx *wait_evt_ctx, *tmp;
	int rem;
	u32 evt_type;
	struct nlattr *tmp_attr_event_type;

	printk(KERN_INFO "%s,%d LOOKING FOR WAIT CONTEXT for ifindex %d\n",
	       __func__, __LINE__, ifindex);

	/* already have spin_lock at this point */
	list_for_each_entry_safe(wait_evt_ctx, tmp,
				 &icrdmadbg_evt_ctx_wait_list, node) {
		if (wait_evt_ctx->ifindex == ifindex) {
			printk(KERN_INFO "%s,%d UPDATING WAIT CONTEXT for \
				ifindex %d from PID %d after STOP\n", __func__, __LINE__, ifindex, pid);

			if (wait_evt_ctx->client_ref_cnt >= 1)
				wait_evt_ctx->client_ref_cnt--;

			printk(KERN_INFO
			       "%s,%d %d client(s) waiting for ifindex %d\n",
			       __func__, __LINE__, wait_evt_ctx->client_ref_cnt,
			       ifindex);

			nla_for_each_nested(tmp_attr_event_type,
					    attrs_nested, rem) {
				evt_type = nla_get_u32(tmp_attr_event_type);
				if (evt_type < IRDMADBG_EVENT_TYPE_COUNT &&
				    wait_evt_ctx->evt_ref_cnt[evt_type] >= 1)
					wait_evt_ctx->evt_ref_cnt[evt_type] -=
					    1;
			}

			if (wait_evt_ctx->client_ref_cnt == 0) {
				printk(KERN_INFO
				       "%s,%d DELETING WAIT CONTEXT for ifindex %d\n",
				       __func__, __LINE__, ifindex);

				list_del(&(wait_evt_ctx->node));
				IW_KFREE(wait_evt_ctx);
			}

			break;
		}
	}

	return IRDMA_SUCCESS;
}

/*
 * Configures event subscription for given ifindex
 * and manages ghost contexts for interfaces that
 * are not present in the system yet
 */
static int iwdbg_nl_cmd_event_cfg(struct sk_buff *msg, struct genl_info *info)
{
	struct nlattr *attrs_nested[IRDMADBG_ATTR_EVENT_CFG_COUNT];
	struct nlattr *tmp_attr_event_type;
	struct irdma_device *iwdev;
	struct icrdmadbg_evt_ctx *wait_evt_ctx;
	struct net *net;
	int ret, rem, sign, ifindex;
	u32 monitor_state;

	if (!info->attrs[IRDMADBG_ATTR_IFINDEX]) {
		printk(KERN_INFO
		       "%s,%d info->attrs[IRDMADBG_ATTR_IFINDEX] problem\n",
		       __func__, __LINE__);
		goto err;
	}

	if (!info->attrs[IRDMADBG_ATTR_EVENT_CFG]) {
		printk(KERN_INFO
		       "%s,%d info->attrs[IRDMADBG_ATTR_EVENT_CFG] problem\n",
		       __func__, __LINE__);
		goto err;
	}

	ret = nla_parse_nested(attrs_nested, IRDMADBG_ATTR_EVENT_CFG_MAX,
			       info->attrs[IRDMADBG_ATTR_EVENT_CFG],
			       icrdmadbg_nl_event_cfg_policy);

	if (ret != 0) {
		printk(KERN_INFO "%s,%d ret = nla_parse_nested for \
			IRDMADBG_ATTR_EVENT_CFG_MAX:%d problem\n", __func__, __LINE__, ret);
		goto err;
	}

	ifindex = nla_get_u32(info->attrs[IRDMADBG_ATTR_IFINDEX]);

	monitor_state =
	    nla_get_u32(attrs_nested[IRDMADBG_ATTR_EVENT_CFG_MONITOR_STATE]);

	if (monitor_state == IRDMADBG_EVENT_MONITOR_START) {
		printk(KERN_INFO
		       "%s,%d IRDMADBG_EVENT_MONITOR_START for pid:%d\n",
		       __func__, __LINE__, info->snd_pid);
		sign = 1;
	}
	if (monitor_state == IRDMADBG_EVENT_MONITOR_STOP) {
		printk(KERN_INFO
		       "%s,%d IRDMADBG_EVENT_MONITOR_STOP for pid:%d\n",
		       __func__, __LINE__, info->snd_pid);
		sign = -1;
	}
	/* allocate memory in case we need it later */
	/* need to do it ouside of spinloc */
	wait_evt_ctx = kzalloc(sizeof(*wait_evt_ctx), GFP_KERNEL);

	if (!wait_evt_ctx) {
		printk(KERN_ERR
		       "%s:%d wait_evt_ctx Memory allocation failed!\n",
		       __func__, __LINE__);
		return -ENOMEM;
	}

	spin_lock(&icrdmadbg_evt_ctx_wait_list_lock);
	iwdev = iwdbg_get_iwdev_by_index(ifindex);

	if (!iwdev && (monitor_state == IRDMADBG_EVENT_MONITOR_START)) {
		ret = iwdbg_add_monitor_to_wait_evt_ctx(wait_evt_ctx, ifindex,
							attrs_nested
							[IRDMADBG_ATTR_EVENT_CFG_LIST_EVT_TYPE],
							info->snd_pid);

		spin_unlock(&icrdmadbg_evt_ctx_wait_list_lock);
#ifdef IW_DEBUG_PRINT

		ret = print_event_subscriptions_cnt(wait_evt_ctx);
#endif				/* IW_DEBUG_PRINT */

		return ret;
	}

	if (!iwdev && (monitor_state == IRDMADBG_EVENT_MONITOR_STOP)) {
		ret = iwdbg_del_monitor_from_wait_evt_ctx(ifindex,
							  attrs_nested
							  [IRDMADBG_ATTR_EVENT_CFG_LIST_EVT_TYPE],
							  info->snd_pid);

		spin_unlock(&icrdmadbg_evt_ctx_wait_list_lock);
#ifdef IW_DEBUG_PRINT

		print_event_subscriptions_cnt(iwdev->event_ctx);
#endif				/* IW_DEBUG_PRINT */

		return ret;
	}

	iwdev->event_ctx->client_ref_cnt += sign;

	nla_for_each_nested(tmp_attr_event_type,
			    attrs_nested[IRDMADBG_ATTR_EVENT_CFG_LIST_EVT_TYPE],
			    rem) {
		switch (nla_get_u32(tmp_attr_event_type)) {
		case IRDMADBG_EVENT_TYPE_AEQ:
			iwdev->event_ctx->
			    evt_ref_cnt[IRDMADBG_EVENT_TYPE_AEQ] += sign;
			break;
		case IRDMADBG_EVENT_TYPE_AQ:
			iwdev->event_ctx->evt_ref_cnt[IRDMADBG_EVENT_TYPE_AQ] +=
			    sign;
			break;
		case IRDMADBG_EVENT_TYPE_CEQ_ALL:
			iwdev->event_ctx->
			    evt_ref_cnt[IRDMADBG_EVENT_TYPE_CEQ_ALL] += sign;
			break;
		case IRDMADBG_EVENT_TYPE_CEQ_ERR:
			iwdev->event_ctx->
			    evt_ref_cnt[IRDMADBG_EVENT_TYPE_CEQ_ERR] += sign;
			break;
		case IRDMADBG_EVENT_TYPE_CQ0_ALL:
			iwdev->event_ctx->
			    evt_ref_cnt[IRDMADBG_EVENT_TYPE_CQ0_ALL] += sign;
			break;
		case IRDMADBG_EVENT_TYPE_CQ0_ERR:
			iwdev->event_ctx->
			    evt_ref_cnt[IRDMADBG_EVENT_TYPE_CQ0_ERR] += sign;
			break;
		default:
			printk(KERN_INFO "%s,%d event=%d NOT supported\n",
			       __func__, __LINE__,
			       nla_get_u32(tmp_attr_event_type));
		}
	}
	printk(KERN_INFO "%s,%d process events = %d\n", __func__, __LINE__,
	       iwdev->event_ctx->client_ref_cnt);
#ifdef IW_DEBUG_PRINT

	print_event_subscriptions_cnt(iwdev->event_ctx);
#endif				/* IW_DEBUG_PRINT */

	net = dev_net(iwdev->netdev);

	spin_unlock(&icrdmadbg_evt_ctx_wait_list_lock);

	if (monitor_state == IRDMADBG_EVENT_MONITOR_START) {
		ret = icrdmadbg_process_adapater_status_event(net,
							     IRDMADBG_EVENT_ADAPTER_STATUS_REGISTER_DEV_EVT_ID,
							     ifindex);
	}

	return IRDMA_SUCCESS;
 err:
	return -EINVAL;
}

/*
 * Processes adapter status event -
 * send current status debug tool
 */
int icrdmadbg_process_adapater_status_event(struct net *net,
					   enum
					   icrdmadbg_event_adapter_status_evt_id
					   event_id, int ifindex)
{
	struct sk_buff *skb;
	struct nlattr *attrs;
	int ret;
	void *hdr;

	if (!net) {
		printk(KERN_ERR "%s,%d net problem\n", __func__, __LINE__);
		return -1;	/* skip event indication */
	}
	skb = nlmsg_new(NLMSG_GOODSIZE, GFP_KERNEL);
	hdr = genlmsg_put(skb, 0, 0, &icrdmadbg_nl_family, 0, event_id);

	if (!hdr) {
		printk(KERN_ERR "%s,%d resp hdr roblem !\n",
		       __func__, __LINE__);

		return -EINVAL;
	}

	ret = nla_put_u32(skb, IRDMADBG_ATTR_IFINDEX, ifindex);
	if (ret)
		goto err_put;

	attrs = nla_nest_start(skb, IRDMADBG_ATTR_EVENT);
	if (!attrs) {
		printk(KERN_INFO "%s,%d attrs problem\n", __func__, __LINE__);
		return -EINVAL;
	}

	ret = nla_put_u32(skb, IRDMADBG_ATTR_EVENT_TYPE,
			  IRDMADBG_EVENT_TYPE_ADAPTER_STATUS);
	if (ret)
		goto err_put;

	ret = nla_put_u32(skb, IRDMADBG_ATTR_EVENT_ID, event_id);
	if (ret)
		goto err_put;

	nla_nest_end(skb, attrs);
	genlmsg_end(skb, hdr);
	ret = genlmsg_multicast_netns(net, skb, 0,
				      icrdmadbg_events_mc_group.id, GFP_KERNEL);

	if (ret < 0)
		printk(KERN_ERR "%s,%d  genlmsg_multicast ret=%d\n",
		       __func__, __LINE__, ret);

	return ret;

 err_put:
	printk(KERN_ERR "%s,%d nla_put binary data ret=%d\n",
	       __func__, __LINE__, ret);
	nlmsg_free(skb);

	return ret;
}

/*
 * Orocesses CQP event
 *  support for CEQ AQ CQP events  is not enabled
 */
void icrdmadbg_process_cqp_event(struct irdma_device *iwdev,
				struct irdma_cqp_compl_info *cqe_info)
{
	struct sk_buff *skb;
	struct net *net;
	struct nlattr *attrs;
	struct icrdmadbg_cqp_event_info cqp_event_info;
	enum icrdmadbg_event_cq0_evt_id evt_id;
	enum icrdmadbg_event_type event_type;
	int ret;
	void *hdr;

	if (!iwdev) {
		printk(KERN_ERR "%s,%d struct irdma_device iwdev problem\n",
		       __func__, __LINE__);
		return;		/* skip event indication */
	}

	if ((!iwdev->event_ctx->evt_ref_cnt[IRDMADBG_EVENT_TYPE_CQ0_ALL]) &&
	    (!iwdev->event_ctx->evt_ref_cnt[IRDMADBG_EVENT_TYPE_CQ0_ERR]))
		return;

	if ((!iwdev->event_ctx->evt_ref_cnt[IRDMADBG_EVENT_TYPE_CQ0_ALL]) &&
	    (!cqe_info->error))
		return;

	net = dev_net(iwdev->netdev);
	if (!net) {
		printk(KERN_ERR "%s,%d struct net problem\n",
		       __func__, __LINE__);
		return;		/* skip event indication */
	}
	event_type = cqe_info->error ?
	    IRDMADBG_EVENT_TYPE_CQ0_ERR : IRDMADBG_EVENT_TYPE_CQ0_ALL;

	/* information below is redundant, but follows HLD design */
	evt_id = cqe_info->error ?
	    IRDMADBG_EVENT_CQ0_ERROR_EVT_ID :
	    IRDMADBG_EVENT_CQ0_COMPLETED_EVT_ID;

	cqp_event_info.maj_err_code = cqe_info->maj_err_code;
	cqp_event_info.min_err_code = cqe_info->min_err_code;
	cqp_event_info.op_code = cqe_info->op_code;
	cqp_event_info.op_ret_val = cqe_info->op_ret_val;
	cqp_event_info.error = cqe_info->error;

	skb = nlmsg_new(NLMSG_GOODSIZE, GFP_KERNEL);
	if (!skb) {
		printk(KERN_ERR "%s,%d resp skb roblem !\n",
		       __func__, __LINE__);

		return;
	}

	hdr = genlmsg_put(skb, 0, 0, &icrdmadbg_nl_family, 0, event_type);

	if (!hdr) {
		printk(KERN_ERR "%s,%d resp hdr roblem !\n",
		       __func__, __LINE__);

		return;
	}

	ret = nla_put_u32(skb, IRDMADBG_ATTR_IFINDEX, iwdev->netdev->ifindex);
	if (ret)
		goto err_put;

	attrs = nla_nest_start(skb, IRDMADBG_ATTR_EVENT);
	if (!attrs) {
		printk(KERN_ERR "%s,%d attrs problem\n", __func__, __LINE__);
		return -EINVAL;
	}
	ret = nla_put_u32(skb, IRDMADBG_ATTR_EVENT_TYPE, event_type);
	if (ret)
		goto err_put;

	ret = nla_put_u32(skb, IRDMADBG_ATTR_EVENT_ID, evt_id);
	if (ret)
		goto err_put;

	ret = nla_put(skb, IRDMADBG_ATTR_EVENT_DATA,
		      sizeof(cqp_event_info), &cqp_event_info);
	if (ret)
		goto err_put;

	nla_nest_end(skb, attrs);
	ret = genlmsg_end(skb, hdr);
	ret = genlmsg_multicast_netns(net, skb, 0, icrdmadbg_events_mc_group.id,
				      GFP_KERNEL);
	if (ret)
		printk(KERN_ERR "%s,%d genlmsg_multicast ret=%d problem\n",
		       __func__, __LINE__, ret);

	return;

 err_put:

	printk(KERN_ERR "%s,%d nla_put binary data ret=%d\n",
	       __func__, __LINE__, ret);
	nlmsg_free(skb);

	return;
}

/*
 * ops functions are now defined and we can use them
 */
static struct genl_ops icrdmadbg_nl_ops[] = {
	{
	 .cmd = IRDMADBG_CMD_REG,
	 .doit = iwdbg_nl_cmd_reg,
	 .policy = icrdmadbg_nl_policy,
	 .flags = GENL_ADMIN_PERM,
	 },
	{
	 .cmd = IRDMADBG_CMD_HMC,
	 .dumpit = iwdbg_nl_cmd_hmc_dump,
	 .policy = icrdmadbg_nl_policy,
	 .flags = GENL_ADMIN_PERM,
	 },
	{
	 .cmd = IRDMADBG_CMD_EVENT_CFG,
	 .doit = iwdbg_nl_cmd_event_cfg,
	 .policy = icrdmadbg_nl_policy,
	 .flags = GENL_ADMIN_PERM,
	 },
};

/*
 * Initialization / deinitialization functions
 */
/* icrdmadbg_register_device adds iWARP device to
 * tracking list
 */
int icrdmadbg_register_device(struct irdma_device *iwdev)
{
	struct icrdmadbg_device *new_icrdmadbg_device;
	struct icrdmadbg_evt_ctx *tmp, *wait_evt_ctx;
	struct net *net;
	int ifindex;

	if (!iwdev)
		return -EINVAL;

	iwdev->event_ctx =
	    kzalloc(sizeof(*new_icrdmadbg_device), GFP_KERNEL);

	if (!iwdev->event_ctx) {
		printk(KERN_ERR "%s,%d iwdev->event_ctx problem\n",
		       __func__, __LINE__);

		return -ENOMEM;
	}

	new_icrdmadbg_device = kzalloc(sizeof(*new_icrdmadbg_device), GFP_KERNEL);

	if (!new_icrdmadbg_device) {
		printk(KERN_ERR
		       "%s:%d new_icrdmadbg_device memory allocation failed!\n",
		       __func__, __LINE__);
		IW_KFREE(iwdev->event_ctx);
		iwdev->event_ctx = NULL;

		return -ENOMEM;
	}

	/* check if ghost ctxs are waiting for this device */

	printk(KERN_INFO "%s,%d LOOKING GHOST CONTEXT for ifindex %d  on %s\n",
	       __func__, __LINE__, iwdev->netdev->ifindex, iwdev->netdev->name);

	spin_lock(&icrdmadbg_evt_ctx_wait_list_lock);
	list_for_each_entry_safe(wait_evt_ctx, tmp,
				 &icrdmadbg_evt_ctx_wait_list, node) {
		if (wait_evt_ctx->ifindex == iwdev->netdev->ifindex) {
			printk(KERN_INFO
			       "%s,%d UPDATING CONTEXT FOR ifindex %d on %s\n",
			       __func__, __LINE__, iwdev->netdev->ifindex,
			       iwdev->netdev->name);

			memcpy(iwdev->event_ctx, wait_evt_ctx,
			       sizeof(*wait_evt_ctx));

#ifdef IW_DEBUG_PRINT
			print_event_subscriptions_cnt(iwdev->event_ctx);
#endif				/* IW_DEBUG_PRINT */
			list_del(&(wait_evt_ctx->node));
			IW_KFREE(wait_evt_ctx);
		}
	}

	new_icrdmadbg_device->iwdev = iwdev;
	spin_lock(&icrdmadbg_iw_dev_list_lock);
	list_add(&(new_icrdmadbg_device->node), &icrdmadbg_iw_dev_list);
	spin_unlock(&icrdmadbg_iw_dev_list_lock);

	net = dev_net(iwdev->netdev);
	ifindex = iwdev->netdev->ifindex;
	spin_unlock(&icrdmadbg_evt_ctx_wait_list_lock);

	if (iwdev->event_ctx->client_ref_cnt)
		icrdmadbg_process_adapater_status_event(net,
						       IRDMADBG_EVENT_ADAPTER_STATUS_REGISTER_DEV_EVT_ID,
						       ifindex);

	return IRDMA_SUCCESS;
}

/* icrdmadbg_unregister_device removes iWARP device from
 * debug tracking list
 */
void icrdmadbg_unregister_device(struct irdma_device *iwdev)
{
	struct icrdmadbg_device *tmp_iwdbg_device, *tmp;
	struct net *net;
	int ifindex;

	spin_lock(&icrdmadbg_evt_ctx_wait_list_lock);
	list_for_each_entry_safe(tmp_iwdbg_device, tmp, &icrdmadbg_iw_dev_list,
				 node) {
		if (iwdev == tmp_iwdbg_device->iwdev) {
			spin_lock(&icrdmadbg_iw_dev_list_lock);
			list_del(&(tmp_iwdbg_device->node));
			spin_unlock(&icrdmadbg_iw_dev_list_lock);
			IW_KFREE(tmp_iwdbg_device);
			break;
		}
	}
	net = dev_net(iwdev->netdev);
	ifindex = iwdev->netdev->ifindex;

	spin_unlock(&icrdmadbg_evt_ctx_wait_list_lock);

	if (iwdev->event_ctx->client_ref_cnt)
		icrdmadbg_process_adapater_status_event(net,
						       IRDMADBG_EVENT_ADAPTER_STATUS_UNREGISTER_DEV_EVT_ID,
						       ifindex);

	IW_KFREE(iwdev->event_ctx);
	iwdev->event_ctx = NULL;
}

/* icrdmadbg_init this function
 * registers netlink family for HW debugging tools
 * requests handling
 */
int icrdmadbg_init(void)
{
	int ret;

	ret = genl_register_family_with_ops(&icrdmadbg_nl_family,
					    icrdmadbg_nl_ops,
					    ARRAY_SIZE(icrdmadbg_nl_ops));
	if (ret != 0) {
		printk(KERN_ERR "%s,%d HW debug initialization problem\n",
		       __func__, __LINE__);
		goto register_family_out;
	}

	ret = genl_register_mc_group(&icrdmadbg_nl_family,
				     &icrdmadbg_events_mc_group);
	if (ret != 0) {
		printk(KERN_ERR
		       "%s,%d HW debug mc_group initialization problem\n",
		       __func__, __LINE__);
		goto register_mc_group_out;
	}

	return IRDMA_SUCCESS;

 register_mc_group_out:
	genl_unregister_family(&icrdmadbg_nl_family);

 register_family_out:
	return ret;
}

/* icrdmadbg_deinit this function
 * de-register netlink family for HW debugging tools
 * requests handling
 */
void icrdmadbg_deinit(void)
{
	genl_unregister_mc_group(&icrdmadbg_nl_family,
				 &icrdmadbg_events_mc_group);
	genl_unregister_family(&icrdmadbg_nl_family);
}
