/* Copyright (c) 2016 - 2023 Intel Corporation */
#ifndef LICENSE_REPLACED
/* For licensing information, see the file 'LICENSE' in the root folder */
#endif /* LICENSE_REPLACED */
#include "msg.h"
#include <status.h>
#include "defs.h"

/**
 * set_field - Set field value in u32 buffer.
 */
void iw_vf_set_field(u32 * buf, u64 field_val, u8 shift, u64 mask)
{
	ASSERT(((field_val << shift) & (~mask)) == 0);
	*buf = ((*buf) & (~mask)) | ((field_val << shift) & mask);
}

/**
 * pack_iw_msg - pack uda/iwarp AQ message to message buffer
 * @iw_msg_in: iwarp_msg structure
 * @msg_out: output message buffer
 * @msg_len: output message length
 */
int pack_iw_msg(struct iwarp_msg *iw_msg_in, u32 *msg_out, u16 *msg_len)
{
	int ret = IRDMA_SUCCESS;
	u8 i;

	ASSERT(msg_out != NULL);
	ASSERT(msg_len != NULL);
	ASSERT(iw_msg_in != NULL);

	switch (iw_msg_in->msg_id) {
	case IRDMA_VF_ALLOC_ALL_HMC_OBJ:
		if (*msg_len < ALLOC_ALL_HMC_OBJ_MSG_LEN)
			return IRDMA_ERR_PARAM;

		for (i = 0; i < I40E_HMC_IW_MAX; i++) {
			SET_ALLOC_MSG(msg_out, i, OBJECT_TYPE,
				      iw_msg_in->mtype.alloc_obj[i].rsrc_type);

			SET_ALLOC_MSG(msg_out, i, HMC_ENTRY_TYPE,
				      iw_msg_in->mtype.alloc_obj[i].entry_type);

			SET_ALLOC_MSG(msg_out, i, OBJECT_INDEX,
				      iw_msg_in->mtype.alloc_obj[i].start_idx);
			SET_ALLOC_MSG(msg_out, i, OBJECT_CNT,
				      iw_msg_in->mtype.alloc_obj[i].count);

			/* hmc address offset */
			SET_ALLOC_MSG(msg_out, i, HIGH_ADDR_OFF,
				      iw_msg_in->mtype.alloc_obj[i].
				      addr_offset);
			SET_ALLOC_MSG(msg_out, i, LOW_ADDR_OFF,
				      iw_msg_in->mtype.alloc_obj[i].
				      addr_offset);
		}
		iw_msg_in->msg_len = ALLOC_ALL_HMC_OBJ_MSG_LEN;
		break;
	case IRDMA_VF_FREE_ALL_HMC_OBJ:
		if (*msg_len < FREE_ALL_HMC_OBJ_MSG_LEN)
			return IRDMA_ERR_PARAM;

		for (i = 0; i < I40E_HMC_IW_MAX; i++) {
			SET_FREE_MSG(msg_out, i, OBJECT_TYPE,
				     iw_msg_in->mtype.alloc_obj[i].rsrc_type);
			SET_FREE_MSG(msg_out, i, OBJECT_INDEX,
				     iw_msg_in->mtype.alloc_obj[i].start_idx);
			SET_FREE_MSG(msg_out, i, OBJECT_CNT,
				     iw_msg_in->mtype.alloc_obj[i].count);
		}
		iw_msg_in->msg_len = FREE_ALL_HMC_OBJ_MSG_LEN;
		break;
	case IRDMA_VF_ALLOC_HMC_FNID:
		/* this message sends only header to PF */
		if (*msg_len < ALLOC_HMC_FNID_MSG_LEN)
			return IRDMA_ERR_PARAM;

		SET_FIELD(msg_out, HMC_FNID,
			  iw_msg_in->mtype.hmc_fnid_obj.hmc_fn_id);
		iw_msg_in->msg_len = ALLOC_HMC_FNID_MSG_LEN;
		break;
	case IRDMA_VF_CHANNEL_VER:
		/* this message sends channel version to PF */
		SET_FIELD(msg_out, CHANNEL_VER, OPAGUE_CHANNEL_VERSION);
		iw_msg_in->msg_len = CHANNEL_VER_MSG_LEN;
		break;
	default:
		return IRDMA_ERR_PARAM;
	}

	/* set message header */
	msg_out[0] = FIELD_PREP(MBOX_ID, iw_msg_in->mbox_id) |
		     FIELD_PREP(MSG_ID, iw_msg_in->msg_id);
	msg_out[1] = FIELD_PREP(RESP_CODE, iw_msg_in->responce_code)
	    | FIELD_PREP(SEQ_NUM, iw_msg_in->seq_num)
	    | FIELD_PREP(MSG_LEN, iw_msg_in->msg_len);

	/* msg length in bytes */
	*msg_len = iw_msg_in->msg_len * sizeof(u32);
	return ret;
}

/**
 * parse_iw_msg - parse message buffer and set iwarp/uda message structure
 * @iwarp_msg_rcv: iwarp_msg structure
 * @msg_in: input message buffer
 * @msg_len: input message length
 * When error code is returned it is should be information for L2 driver
 * to not send the ACK answer causing timeout error on other side.
 */
int parse_iw_msg(struct iwarp_msg *iwarp_msg_rcv, u32 *msg_in, u16 msg_len)
{
	int ret = IRDMA_SUCCESS;
	u32 i;

	if (msg_in == NULL)
		return IRDMA_ERR_PARAM;

	/* message length in 32bits word */
	msg_len = msg_len / 4;

	if (msg_len < HEADER_LEN)
		return IRDMA_ERR_PARAM;

	memset(iwarp_msg_rcv, 0, sizeof(*iwarp_msg_rcv));
	iwarp_msg_rcv->msg_id = FIELD_GET(MSG_ID, msg_in[0]);
	iwarp_msg_rcv->mbox_id = FIELD_GET(MBOX_ID, msg_in[0]);
	iwarp_msg_rcv->responce_code = FIELD_GET(RESP_CODE, msg_in[1]);
	iwarp_msg_rcv->seq_num = FIELD_GET(SEQ_NUM, msg_in[1]);
	iwarp_msg_rcv->msg_len = FIELD_GET(MSG_LEN, msg_in[1]);

	if (iwarp_msg_rcv->msg_len != msg_len)
		return IRDMA_ERR_PARAM;

	switch (iwarp_msg_rcv->msg_id) {
	case IRDMA_VF_ALLOC_ALL_HMC_OBJ:
		{
			for (i = 0; i < I40E_HMC_IW_MAX; i++) {
				iwarp_msg_rcv->mtype.alloc_obj[i].rsrc_type =
				    FIELD_GET(OBJECT_TYPE,
					      GET_ALLOC_MSG(msg_in, i, OBJECT_TYPE));
				iwarp_msg_rcv->mtype.alloc_obj[i].entry_type =
				    FIELD_GET(HMC_ENTRY_TYPE,
					      GET_ALLOC_MSG(msg_in, i, HMC_ENTRY_TYPE));
				iwarp_msg_rcv->mtype.alloc_obj[i].start_idx =
				    FIELD_GET(OBJECT_INDEX,
					      GET_ALLOC_MSG(msg_in, i, OBJECT_INDEX));
				iwarp_msg_rcv->mtype.alloc_obj[i].count =
				    FIELD_GET(OBJECT_CNT,
					      GET_ALLOC_MSG(msg_in, i, OBJECT_CNT));
				/* hmc address offset */
				iwarp_msg_rcv->mtype.alloc_obj[i].addr_offset =
				    FIELD_GET(LOW_ADDR_OFF,
					      GET_ALLOC_MSG(msg_in, i, LOW_ADDR_OFF))
				    +
				    FIELD_GET(HIGH_ADDR_OFF,
					      GET_ALLOC_MSG(msg_in, i, HIGH_ADDR_OFF));
			}
			/* set responce code as OK */
			iwarp_msg_rcv->responce_code = IRDMA_MSG_OK;
			break;
		}

	case IRDMA_VF_FREE_ALL_HMC_OBJ:
		{
			for (i = 0; i < I40E_HMC_IW_MAX; i++) {
				iwarp_msg_rcv->mtype.free_obj[i].rsrc_type =
				    FIELD_GET(OBJECT_TYPE,
					      GET_FREE_MSG(msg_in, i, OBJECT_TYPE));
				iwarp_msg_rcv->mtype.free_obj[i].start_idx =
				    FIELD_GET(OBJECT_INDEX,
					      GET_FREE_MSG(msg_in, i, OBJECT_INDEX));
				iwarp_msg_rcv->mtype.free_obj[i].count =
				    FIELD_GET(OBJECT_CNT,
					      GET_FREE_MSG(msg_in, i, OBJECT_CNT));
			}
			/* set responce code as OK */
			iwarp_msg_rcv->responce_code = IRDMA_MSG_OK;
			break;
		}

	case IRDMA_VF_ALLOC_HMC_FNID:
		{
			iwarp_msg_rcv->mtype.hmc_fnid_obj.hmc_fn_id =
			    FIELD_GET(HMC_FNID, GET_FIELD(msg_in, HMC_FNID));
			/* set responce code as OK */
			iwarp_msg_rcv->responce_code = IRDMA_MSG_OK;
			break;
		}

	case IRDMA_VF_CHANNEL_VER:
		{
			iwarp_msg_rcv->mtype.channel_ver_rsp.version =
			    FIELD_GET(CHANNEL_VER,
				      GET_FIELD(msg_in, CHANNEL_VER));
			iwarp_msg_rcv->responce_code = IRDMA_MSG_OK;
			break;
		}
	default:
		/* set responce code as ERROR */
		iwarp_msg_rcv->responce_code = IRDMA_MSG_ERROR;
		return IRDMA_ERR_PARAM;
	}
	return ret;
}
