
#define CHECKS_AEQ_BAD_RING_HEAD          0x1
#define CHECKS_AEQ_BAD_RING_TAIL          0x2
#define CHECKS_AEQ_BAD_RING_TAIL_POLARITY 0x4
#define CHECKS_AEQ_MOVE_TAIL_OVERFLOW     0x8
#define CHECKS_AEQ_MOVE_HEAD_OVERFLOW     0x10

class Ring {
	u32 ring_tail;
	u32 ring_head;
	u32 ring_size;
	u32 avail_entries;
	u32 entries_filled;
	u8 tail_polarity;
	u8 head_polarity;
	// err format:
	// bit 0: tail move overflow (beyond head)
	// bit 1: head move overflow (beyond tail)
	u8 err;
	u32 checks;
public:
	virtual void init_entries_filled(void)
	{
		entries_filled = (ring_head + ring_size - ring_tail) % ring_size;
	}

	virtual void init(u32 tail, u32 head, u32 size, u8 t_polarity, u8 h_polarity)
	{
		ring_tail = tail;
		ring_head = head;
		ring_size = size;
		tail_polarity = t_polarity;
		head_polarity = h_polarity;
		init_entries_filled();
		err = 0;
		checks = 0;
	}

	virtual void rand_init(void)
	{
		ring_size = rand();
		if (ring_size < 10) {
			ring_size = 10;
		}
		ring_tail = rand() % ring_size;
		ring_head = rand() % ring_size;
		head_polarity = rand() % 2;
		tail_polarity = head_polarity;
		if (ring_head < ring_tail)
			tail_polarity ^= 1;
		err = 0;
		checks = 0;
	}

	virtual bool move_tail(u32 moves)
	{
		u32 prev_ring_tail = ring_tail;

		if (moves > (entries_filled)) {
			err = 0x1;
			return false;
		}
		ring_tail = (ring_tail + moves) % ring_size;
		if (prev_ring_tail > ring_tail) {
			tail_polarity ^= 1;
			tail_polarity &= 0x1;
		}
		return true;
	}

	virtual bool move_head(u32 moves)
	{
		u32 prev_ring_head = ring_head;
		if (moves > ring_size - 1 - entries_filled) {
			err = 0x2;
			return false;
		}
		ring_head = (ring_head + moves) % ring_size;
		if (prev_ring_head > ring_head) {
			head_polarity ^= 1;
			tail_polarity &= 0x1;
		}
		err = 0;
		init_entries_filled();
		return true;
	}

	virtual void clear_err(void)
	{
		err = 0;
	}


	virtual void print_checks(struct irdma_ring *ring, u8 polarity)
	{
		if (checks & CHECKS_AEQ_BAD_RING_HEAD)
			printf("ERROR: BAD RING HEAD - ring->head (0x%x) != ring_head (0x%x)\n",ring->head, ring_head);
		if (checks & CHECKS_AEQ_BAD_RING_TAIL)
			printf("ERROR: BAD RING TAIL - ring->tail (0x%x) != ring_tail (0x%x)\n",ring->tail, ring_tail);
		if (checks & CHECKS_AEQ_BAD_RING_TAIL_POLARITY)
			printf("ERROR: BAD RING TAIL POLARITY - polarity (0x%x) != tail_polarity (0x%x)\n",polarity & 0x1, tail_polarity);
		if (checks & CHECKS_AEQ_MOVE_TAIL_OVERFLOW)
			printf("ERROR: MOVE TAIL OVERFLOW\n");
		if (checks & CHECKS_AEQ_MOVE_HEAD_OVERFLOW)
			printf("ERROR: MOVE HEAD OVERFLOW\n");
	}

	virtual bool check_ring(struct irdma_ring *ring, u8 polarity)
	{
		if (ring->head != ring_head)
			checks |= CHECKS_AEQ_BAD_RING_HEAD;
		if (ring->tail != ring_tail)
			checks |= CHECKS_AEQ_BAD_RING_TAIL;
		if ((polarity & 0x1) != tail_polarity)
			checks |= CHECKS_AEQ_BAD_RING_TAIL_POLARITY;

		if(checks) {
			print_checks(ring, polarity);
			return false;
		}
		return true;
	}

	virtual void get_entry(u8 polarity)
	{
		if (polarity == tail_polarity)
			if (!move_tail(1))
				checks |= CHECKS_AEQ_MOVE_TAIL_OVERFLOW;
	}

	virtual bool add_entry(u8 polarity)
	{
		if (polarity == head_polarity)
			if (!move_head(1)) {
				checks |= CHECKS_AEQ_MOVE_HEAD_OVERFLOW;
				return false;
			} else
				return true;
		else
			return false;
	}
};
