#include "wqe.h"

// ------------------ Field Function Definitions -----------------------

void Field::init(std::string i_name, u64 i_data, u64 i_bit_mask,
			 u8 i_bit_shift, u8 i_bit_length,
			 enum permissions i_perm, u16 i_byte_offset)
{
	name = i_name;
	data = i_data,
	bit_mask = i_bit_mask;
	bit_shift = i_bit_shift;
	bit_length = i_bit_length;
	perm = i_perm;
	byte_offset = i_byte_offset;
}

u64 Field::get_masked_data_64(void)
{
	return data & bit_mask;
}

u64 Field::get_masked_shifted_data_64(void)
{
	return (data & bit_mask) << bit_shift;
}

void Field::print_debug_high(void)
{
	if (debug >= DEBUG_HIGH) {
		std::cout << "name (" << name << ") & data(0x" << std::hex <<
		          data << std::dec << "), perm(" << perm << ")" << std::endl;
	}
}

void INLINE_FIELD::init(std::string i_name, u64 i_data, u64 i_bit_mask,
			u8 i_bit_shift, u8 i_bit_length,
			enum permissions i_perm, u16 i_byte_offset)
{
	Field::init(i_name, i_data, i_bit_mask, i_bit_shift, i_bit_length,
		    i_perm, i_byte_offset);

	quanta = i_byte_offset / TB_BYTES_PER_QUANTA_CPK;
	if (i_byte_offset % TB_BYTES_PER_QUANTA_CPK)
		quanta++;
}

void INLINE_FIELD::set_inline_len(u16 len)
{
	inline_len = len;
	set_mask();
}

void INLINE_FIELD::set_mask(void)
{
	u16 used_quanta = 0;

	if (inline_len > 8) {
		used_quanta = inline_len-8 / TB_INLINE_BYTES_PER_QUANTA_CPK;
		if (inline_len % TB_INLINE_BYTES_PER_QUANTA_CPK)
			used_quanta++;
	} else
		used_quanta = 1;
		
	if (quanta <= used_quanta) {
		perm = READWRITE;
		if (used_quanta >= 2) {
			used_quanta--;
		
		} else if (byte_offset == 8) {
			bit_mask = 0xff;
			for (int i=1; i<inline_len; i++)
				bit_mask |= bit_mask << 8 | 0xff;
		}
	} else {
		perm = RSVD;
		bit_mask = 0;
	}
}

// ------------------ WQE Function Definitions -----------------------

//to be overridden by child classes
//void WQE::rand_init(void) {}

//to be overridden by child classes
//void WQE::gen_field_vs_rules(int field_idx, u64 *field_data) {}

u64 WQE::gen_byte_offset(u16 byte)
{
	mem = 0xDEADBEEFFACECAFE;

	if (byte % 8 != 0)
		return 0;	

	mem = 0;
	// searches through all fields
	for (int i=0; i<num_fields; i++) {
		// if field matches a given offset
		//   mem = mem ORed with field data masked and shifted
		if (fields[i].byte_offset == byte) {
			if (debug >= DEBUG_HIGH) {
				printf("TB idx(%d) has offset (%d), ",i , byte);
				fields[i].print_debug_high();
			}
			if (fields[i].perm != RSVD) {
				u64 tmp = fields[i].data;
				gen_field_vs_rules(i, &tmp);
				tmp = (tmp & fields[i].bit_mask) << fields[i].bit_shift;
				mem |= tmp;
			}
		}
	}
	return mem;
}

u16 WQE::min_wqe_words(void)
{
	return 0;
}

bool WQE::compare_64bits(u64 mem_val, u16 byte_off)
{
	if (debug >= DEBUG_LOW) {
		u64 exp;
		bool same;

		exp = gen_byte_offset(byte_off);
		same = mem_val == exp;
		printf("\t%sMEM DEBUG: BYTE(%d) act(0x%llx), exp(0x%llx)\n",
		       same ? "":"!",
		       byte_off, mem_val, exp);
		return same;
	}
	return mem_val == gen_byte_offset(byte_off);
}

bool WQE::compare_wqe_words(u64 *sq_wqe, u16 head, u16 size, u8 num_words)
{
	bool ret = true;
	u16 idx = 0;
	if (debug >= DEBUG_LOW)
		printf("  TB: %s: head(%d), size(%d)\n", __func__, head, size);
	for (int i=0; i<num_words; i++) {
		idx = (head*min_wqe_words()+i)%(size*min_wqe_words());
		ret &= compare_64bits(sq_wqe[idx], i*8);
	}
	return ret;
}

void WQE::write_wr(u64 *sq_wqe, u16 head, u16 size, u8 num_words)
{
	u16 idx = 0;

	for (int i=0; i<num_words; i++) {
		idx = (head*min_wqe_words()+i)%(size*min_wqe_words());
		sq_wqe[idx] = gen_byte_offset(i*8);
	}
}

Field *WQE::find_field_by_name(const std::string& find)
{
	for (int i=0; i<num_fields; i++) {
		if (fields[i].name.compare(find))
			return &fields[i];
	}
	return NULL;
}

u64 WQE::get_valid(void)
{
	return 0;
}

//to be overridden by child classes
//void WQE::randomize(void) {}

// returns 0 if successful
// returns 1 if cannot find field name
int WQE::set_field_by_name(const std::string& find, u64 new_data)
{
	Field *tmp = find_field_by_name(find);
	if (tmp) {
		tmp->data = new_data;
		return 0;
	}
	return 1;
}

Field *WQE::find_field_by_int(int idx)
{
	return &fields[idx];
}

void WQE::set_field_by_int(int idx, u64 new_val)
{
	fields[idx].data = new_val;
}

// ------------------ CQP_SQ Function Definitions -----------------------

u16 CQP_SQ_WR::min_wqe_words(void)
{
	return 8;
}

// ------------------ QP_SQ Function Definitions -----------------------

u16 QP_SQ_WR::min_wqe_words(void)
{
	return 4;
}

