# Copyright (c) 2016 - 2021 Intel Corporation
# For licensing information, see the file 'LICENSE' in the root folder
#
# build.mk
#
# Makefile for generating stripped source for various build types
#
# Strip flags:  if preceded by a -D,	removes #ifdefs but leaves code
#					removes #ifndefs and code
#
#		if preceded by a -U, 	removes #ifdefs and code
#					removes #ifndefs but leaves code
#
# Keep flags:	if not specified	removes all #ifdefs and code
#					removes all #ifndefs but leaves code
#
#		if specified		leaves #ifdef and code
#					leaves #ifndefs and code
#
#		if prefixed with -	removes #ifdefs but leaves code
#					removes #ifndefs and code
#
# Note that strip flags always apply first!
#
# to get verbose output from this process file build with V=1 argument, output
# is put to STDOUT

# pull in kernel compatibility build settings:
#COMPAT_INC = COMPAT/compat.inc
#ifeq (,$(wildcard $(COMPAT_INC)))
#  $(error Cannot find $(COMPAT_INC) for kcompat flags))
#else
#  include COMPAT/compat.inc
#  KEEP_FLAGS := $(COMPAT_KEEP_FLAGS)
#  STRIP_FLAGS := $(COMPAT_STRIP_FLAGS) -DDRIVER_I40E -USV_SUPPORT -UI40E_QV \
#                 -UDBG
#endif

# source file "groups"
CORE_FILES =	cm.c debugfs.c hw.c main.c \
		i40iw_if.c \
		utils.c verbs.c irdma_kcompat.c \
		main.h cm.h trace_cm.h osdep.h \
		trace.c trace.h \
		verbs.h irdma_kcompat.h rhel_kcompat.h suse_kcompat.h linux_kcompat.h \
		distro_ver.h ofed_kcompat.h ubuntu_kcompat.h oracle_kcompat.h configfs.c \
 		i40e_client.h
CORE_FILES_NOCOMMENT = irdma-abi.h
SHARED_FILES =	ctrl.c hmc.c puda.c pble.c uk.c \
		virtchnl.c uda.c icrdma_hw.c i40iw_hw.c ws.c \
		defs.h hmc.h protos.h puda.h pble.h uda.h uda_d.h \
		irdma.h icrdma_hw.h i40iw_hw.h \
		status.h type.h user.h virtchnl.h ws.h
MEV_SHARED_FILES = ig3rdma_hw.c ig3rdma_hw.h ig3rdma_regs.h ig3rdma_regs_apf.h ig3rdma_regs_avf.h

# output directories
PF_DRIVER_DIR ?= ./irdma

CVID_FILE := ../../comp_id_cvid.txt
# Determine if building on SCM server
SCM_SERVER := $(shell if test -f ${CVID_FILE}; then echo YES; else echo NO; fi)

# Determine VERSION of drivers - if VERSION is not specified via the
# environment, generate one.
ifeq (${VERSION},)
  ifneq (${SCM_SERVER},YES)
    ifeq ($(shell git rev-parse --is-inside-work-tree 2>/dev/null),true)
      # If not building via the Component tool, generate a version string using
      # git-describe. Basically, find the nearest tag matching a specific format
      # and then add "-X-ID" where X is the number of commits past the tag, and
      # ID is the sha1 of the current commit. Then, prefix it with -dirty if the
      # working directory is not clean.
      PF_DESCRIBE := $(shell git describe --dirty --tags --match "[[:digit:]]*" || echo "bad")
      ifeq (${PF_DESCRIBE},"bad")
        $(error No tag available for describing PF driver version. Please fetch git tags.)
      endif
      PF_VERSION := ${PF_DESCRIBE}
    else
      $(warning *** Unable to automatically generate VERSION using git-describe.)
      $(warning *** You should build from within a git working tree.)
      $(error Alternatively you might manually set VERSION on the Make command line)
    endif
  else
    PF_VERSION := $(shell grep -m1 Component_Version ${CVID_FILE} | cut -f2 -d":" | tr -d [:space:])
  endif
else
  PF_VERSION := ${VERSION}
endif
# Note: some package building tools (e.g. rpmbuild) use '-' as a separator
# between version number and other portions of the package name (such as
# architecture) so convert any hyphen in the version string to underscore,
# or, if building with the Component tool bail if any exist.
ifneq (${SCM_SERVER},YES)
  PF_VERSION := $(subst -,_,${PF_VERSION})
else
  # DO NOT convert '-' to something else since that will conflict with the
  # automatically embedded version number in the resultant file names.
  ifeq ($(findstring -,${PF_VERSION}),-)
    $(error PF driver version string ${PF_VERSION} must not contain hyphens)
  endif
endif
# Parse out the major, minor, and build numbers (the fw requires them)
PF_VERSION_MAJOR := $(shell echo ${PF_VERSION} | sed 's/^\([0-9]\+\)\.[0-9]\+\.[0-9]\+\(\.[0-9]\+\)*\(_rc[0-9]\+\)\?\(_[0-9]\+_[a-z0-9]\{8\}\)\?_\?.*/\1/')
ifeq (${PF_VERSION_MAJOR},${PF_VERSION})
  $(warning Incorrect version format '${PF_VERSION}', setting to default VERSION=0.0.0)
  VERSION = 0.0.0
  PF_VERSION = 0.0.0
  PF_VERSION_MAJOR := $(shell echo ${PF_VERSION} | sed 's/^\([0-9]\+\)\.[0-9]\+\.[0-9]\+\(\.[0-9]\+\)*\(_rc[0-9]\+\)\?\(_[0-9]\+_[a-z0-9]\{8\}\)\?_\?.*/\1/')
endif
PF_VERSION_MINOR := $(shell echo ${PF_VERSION} | sed 's/^[0-9]\+\.\([0-9]\+\)\.[0-9]\+\(\.[0-9]\+\)*\(_rc[0-9]\+\)\?\(_[0-9]\+_[a-z0-9]\{8\}\)\?_\?.*/\1/')
PF_VERSION_BUILD := $(shell echo ${PF_VERSION} | sed 's/^[0-9]\+\.[0-9]\+\.\([0-9]\+\)\(\.[0-9]\+\)*\(_rc[0-9]\+\)\?\(_[0-9]\+_[a-z0-9]\{8\}\)\?_\?.*/\1/')
# Parse out the major, minor, and build numbers (the fw requires them)
ifeq ($(filter KERNEL,${BUILD}),)
PF_VERSION_DESC  := $(shell echo ${PF_VERSION} | sed 's/^[0-9]\+\.[0-9]\+\.[0-9]\+\(\.[0-9]\+\)*\(_rc[0-9]\+\)\?\(_[0-9]\+_[a-z0-9]\{8\}\)\?_\?.*/\1\2\3/')
endif

KEEP_FLAGS +=	IRDMA_CM_H \
		IRDMA_DEFS_H \
		IRDMA_TYPE_H \
		IRDMA_USER_H \
		IRDMA_VIRTCHNL_H \
		IRDMA_HMC_H \
		IRDMA_MAIN_H \
		IRDMA_OSDEP_H \
		IRDMA_PUDA_H \
		IRDMA_REGISTER_H \
		IRDMA_STATUS_H \
		IRDMA_VF_H \
		IRDMA_ABI_H \
		IRDMA_VERBS_H \
		IRDMA_DBG_H \
		IRDMA_PROTOS_H \
		IRDMA_PBLE_H \
		IRDMA_UDA_H \
		IRDMA_UDA_D_H \
		I40IW_HW_H \
		IRDMA_H \
		ICRDMA_HW_H \
		IRDMA_WS_H \
		DISTRO_VER_H \
		IRDMA_KCOMPAT_H \
		LINUX_KCOMPAT_H \
		RHEL_KCOMPAT_H \
		SUSE_KCOMPAT_H \
		UBUNTU_KCOMPAT_H \
		OFED_KCOMPAT_H \
		IG3RDMA_HW_H \
		IG3RDMA_REGS_H \
		IG3RDMA_REGS_APF_H \
		IG3RDMA_REGS_AVF_H \
		__TRACE_CM_H \
		TRACE_HEADER_MULTI_READ \
		CREATE_TRACE_POINTS \
		ether_addr_copy \
		eth_zero_addr \
		IB_FW_VERSION_NAME_MAX \
		HAS_IB_SET_DEVICE_OP \
		IB_GET_NETDEV_OP_NOT_DEPRECATED \
		INIT_RDMA_OBJ_SIZE \
		IRDMA_SET_DRIVER_ID \
		ETHER_ADDR_TO_U64 \
		MAKEMASK \
		IN_IFADDR \
		FOR_IFA \
		KERNEL_VERSION \
		LINUX_VERSION_CODE \
		RHEL_RELEASE_CODE \
		RHEL_RELEASE_VERSION \
		SLE_VERSION \
		SLE_VERSION_CODE \
		SLE_LOCALVERSION_CODE \
		CONFIG_SUSE_KERNEL \
		UTS_UBUNTU_RELEASE_ABI \
		UBUNTU_200405 \
		UTS_RELEASE \
		ORACLE_REL \
		ORACLE_KCOMPAT_H \
		CONFIG_IPV6 \
		CONFIG_DEBUG_FS \
		CONFIG_CONFIGFS_FS \
		CONFIG_64BIT \
		CONFIG_DYNAMIC_DEBUG \
		__has_attribute \
		__GCC4_has_attribute___noclone__ \
		__OFED_BUILD__ \
		__OFED_4_8__ \
		OFED_VERSION_CODE \
		IRDMA_ADD_DEL_GID \
		IB_GET_CACHED_GID \
		IB_GET_ETH_SPEED \
		QUERY_GID_ROCE_V2 \
		QUERY_GID_ROCE_V1 \
		ALLOC_HW_STATS_V3 \
		ALLOC_HW_STATS_V2 \
		ALLOC_HW_STATS_V1 \
		GET_HW_STATS_V2 \
		GET_HW_STATS_V1 \
		GET_LINK_LAYER_V2 \
		GET_LINK_LAYER_V1 \
		GET_NETDEV_OP_V2 \
		IW_PORT_IMMUTABLE_V2 \
		IW_PORT_IMMUTABLE_V1 \
		MODIFY_PORT_V2 \
		MODIFY_PORT_V1 \
		QUERY_GID_V2 \
		QUERY_GID_V1 \
		QUERY_PKEY_V2 \
		QUERY_PKEY_V1 \
		QUERY_PORT_V2 \
		QUERY_PORT_V1 \
		ROCE_PORT_IMMUTABLE_V2 \
		ROCE_PORT_IMMUTABLE_V1 \
		CREATE_AH_VER_6 \
		CREATE_AH_VER_5 \
		CREATE_AH_VER_4 \
		CREATE_AH_VER_3 \
		CREATE_AH_VER_2 \
		CREATE_AH_VER_1_1 \
		CREATE_AH_VER_1_2 \
		CREATE_AH_VER_0 \
		ETHER_COPY_VER_1 \
		ETHER_COPY_VER_2 \
		DESTROY_AH_VER_4 \
		DESTROY_AH_VER_3 \
		DESTROY_AH_VER_2 \
		DESTROY_AH_VER_1 \
		ALLOC_UCONTEXT_VER_2 \
		ALLOC_UCONTEXT_VER_1 \
		DEALLOC_UCONTEXT_VER_2 \
		DEALLOC_UCONTEXT_VER_1 \
		ALLOC_PD_VER_3 \
		ALLOC_PD_VER_2 \
		ALLOC_PD_VER_1 \
		IRDMA_DESTROY_CQ_VER_4 \
		IRDMA_DESTROY_CQ_VER_3 \
		IRDMA_DESTROY_CQ_VER_2 \
		IRDMA_DESTROY_CQ_VER_1 \
		IRDMA_DESTROY_SRQ_VER_3 \
		IRDMA_DESTROY_SRQ_VER_2 \
		IRDMA_DESTROY_SRQ_VER_1 \
		IRDMA_ALLOC_MW_VER_2 \
		IRDMA_ALLOC_MW_VER_1 \
		IRDMA_ALLOC_MR_VER_1 \
		IRDMA_ALLOC_MR_VER_0 \
		DEALLOC_PD_VER_4 \
		DEALLOC_PD_VER_3 \
		DEALLOC_PD_VER_2 \
		DEALLOC_PD_VER_1 \
		SET_PCIDEV_PARENT \
		SET_ROCE_CM_INFO_VER_3 \
		SET_ROCE_CM_INFO_VER_2 \
		SET_ROCE_CM_INFO_VER_1 \
		IB_MTU_CONVERSIONS \
		CREATE_QP_VER_2 \
		CREATE_QP_VER_1 \
		DESTROY_QP_VER_2 \
		DESTROY_QP_VER_1 \
		DEREG_MR_VER_2 \
		DEREG_MR_VER_1 \
		CREATE_CQ_VER_3 \
		CREATE_CQ_VER_2 \
		CREATE_CQ_VER_1 \
		COPY_USER_PGADDR_VER_4 \
		COPY_USER_PGADDR_VER_3 \
		COPY_USER_PGADDR_VER_2 \
		COPY_USER_PGADDR_VER_1 \
		GLOBAL_QP_MEM \
		VMA_DATA \
		USE_KMAP \
		UBUNTU_VERSION \
		UBUNTU_VERSION_CODE \
		UBUNTU_2204 \
		UBUNTU_200404 \
		UBUNTU_200403 \
		UBUNTU_200402 \
		UBUNTU_2004 \
		UBUNTU_1804 \
		UVERBS_CMD_MASK \
		USE_QP_ATTRS_STANDARD \
		IB_QP_ATTR_STANDARD_BITS \
		SLES_15_SP_5 \
		SLES_15_SP_4 \
		SLES_15_SP_3 \
		SLES_15_SP_2 \
		SLES_15_SP_1 \
		SLES_15 \
		SLES_12_SP_5 \
		SLES_12_SP_4 \
		SLES_12_SP_3 \
		RHEL_9_2 \
		RHEL_9_1 \
		RHEL_9_0 \
		RHEL_8_8 \
		RHEL_8_7 \
		RHEL_8_6 \
		RHEL_8_5 \
		RHEL_8_4 \
		RHEL_8_3 \
		RHEL_8_2 \
		RHEL_8_1 \
		RHEL_8_0 \
		RHEL_7_6 \
		RHEL_7_5 \
		RHEL_7_4 \
		ALIGN_DOWN \
		REREG_MR_VER_1 \
		REREG_MR_VER_2 \
		ENABLE_DUP_CM_NAME_WA \
		NEED_IDA_ALLOC_MIN_MAX_RANGE_FREE \
		MODULE

# These flags should be preserved for all the targets except the upstream
# builds (BUILD=KERNEL).  For upstream, those flags should be stripped,
# so in case of KERNEL target they must be added to STRIP_FLAGS either
# with -U or -D option.
# For macros like IB_UMEM_GET_Vx, only the one variant (usually the most
# recent one) should be left.
# Thanks to the fact the strip flags are applied first, they will be stripped
# for upstream builds, even though they are in KEEP_FLAGS.
KEEP_FLAGS +=	SET_BEST_PAGE_SZ_V1 \
		SET_BEST_PAGE_SZ_V2 \
		IB_UMEM_GET_V0 \
		IB_UMEM_GET_V1 \
		IB_UMEM_GET_V2 \
		IB_UMEM_GET_V3 \
		IB_IW_PKEY \
		IB_IW_MANDATORY_AH_OP \
		IB_DEALLOC_DRIVER_SUPPORT \
		RDMA_MMAP_DB_SUPPORT \
		FIELD_PREP \
		rdma_umem_for_each_dma_block \
		fallthrough \
		ibdev_dbg \
		refcount_t \
		ALLOC_HW_STATS_STRUCT_V2 \
		ALLOC_HW_STATS_STRUCT_V1 \
		RHEL_7_2 \
		RHEL_7_7 \
		RHEL_7_8 \
		RHEL_7_9 \
		IB_DEV_CAPS_VER_2 \
		NETDEV_TO_IBDEV_SUPPORT \
		IRDMA_IRQ_UPDATE_AFFINITY \
		IRDMA_AUX_GET_SET_DRV_DATA \
		SET_DMABUF

# these are flags to be stripped from ALL builds
STRIP_FLAGS +=  -DEXTERNAL_RELEASE \
		-D__KERNEL__ \
		-DHPM_LINUX \
		-DRDMA_REPORT_SQE_LENGTH \
		-DDCB_SUPPORT \
		-DQOS_SUPPORT \
		-DTERM_SUPPORT \
		-UINTERNAL_ONLY \
		-UHPM_ESXI \
		-UHPM_WINDOWS \
		-UHPM_FBSD \
		-UHPM_FBSD_UPSTREAM \
		-DLICENSE_REPLACED \
		-U__VMKLNX__ \
		-UNDK_DEBUG \
		-UWPP_TRACING \
		-UIRDMA_HW_DBG \
		-UCHECK_PARAMS  \
		-UDEBUG_STATS \
		-UUDA_ENABLED \
		-UENABLE_ALIGNMENT_CHECKS \
		-UCHECK_MAX_MESSAGE_SIZE \
		-UIRDMA_DEBUG \
		-UIRDMA_DEBUG_UK \
		-UMODPARAM_DEBUG \
		-USHARED_DEBUG_VERBOSE \
		-UPOST_WQE_ZERO \
		-ULS_64_1 \
		-USV_DRIVER \
		-UNO_IWARP_CM \
		-UCTRL_ONLY \
		-UCPK_A0_ERRATA_LFC \
		-UCPF_LIB \
		-UKMDK_MOD_QP_BEFORE_RST_EVENT_WA \
		-UKMDK_SQ_COMPL_UNSIGNALED_WA \
		-DCPK_FPK_ORRQ_FULL \
		-UCPK_RDMA_FORCE_READ_FENCE \
		-URDMA_ROCEV2_FORCE_READ_FENCE \
		-URDPU_ASSIST_WA \
		-UCPK_DUMP_TXSCHEDULER_TREE \
		-UCPK_USE_MIN_RQ_WQE_SIZE_WA \
		-UMEV_EMR12_SUPPORT \
		-UCPK_CHECK_MISSING_SQ_CMPL \
		-USCALEUP_REDUCE_MAX_MSG_SIZE \
		-UCPK_CHECK_MISSING_SQ_CMPL \
		-UOFI_ENFORCE_SQ_LENGTH \
		-UESX_NO_CMPL_SUCCESS_FLUSH_SEEN_WA \
		-UESXI_XF_WA \
		-UESXI_Q1_WA \
		-UREFLUSH_DOUBLE_WQE_REPORTING_WA \
		-USKIP_UNSIGNALED_CMPLS \
		-URQ_EMPTY_RETRANS_DBG \
		-UCPK_A0_ERRATA \
		-UDEBUG_SUPPORT_INTERNAL \
		-DAH_REUSE \
		-DDCQCN_CONFIG_SUPPORT \
		-DCONFIG_DEV_LIMITS \
		-DCPK_READFENCE_MITIGATION \
		-DCPK_MARKER_MATCH_TEMP \
		-UIROCE_CM \
		-UVELOCE \
		-UNO_TERMINATE_WA \
		-UUD_QS_ASSIGN_WA \
		-UCLIP_Q1_CNT \
		-DDEVLINK_PORT_CFG_WA \
		-UMEV_MW_BIND_WA \
		-DDISABLE_IWARP_VLAN_0_WA \
		-UWRITE_IMM_ERRATA \
		-UKMODE_PUSH_PAGE\
		-UMEV_PUSH_DBL_64BIT \
		-UMEV_DEBUG_SUPPORT \
		-DIIDC_GET_RDMA_GEN_WA \
		-DNO_PARENT_AUX_DEV_WA \
		-DDUP_CM_NAME_WA \
		-DFPK_SUPPORT \
		-UMPK_SUPPORT

# DEVLINK_PORT_CFG_WA above should be removed once ICE implements irdma devlink options.

# target specific strip/keep flags
STRIP_FLAGS_FPGA = IRDMA_FPGA_SUPPORT

STRIP_FLAGS_MEV = MEV_SUPPORT MEV_HVL OFI_ENFORCE_SQ_LENGTH MEV_IIDC_WA MEV_DEBUG_SUPPORT

STRIP_FLAGS_PF_SUPPORT_VF = VCHNL_SUPPORT

STRIP_FLAGS_B0 = RDPU_ASSIST_WA CPK_TX_SPAD_WA CPK_PREFETCH_CACHE_DISABLE \
	CPK_RDPU_WA CPK_DISABLE_IMP_NACK CPK_CQE_WA \
	IRDMA_FLOW_CONTROL_WARNING CPK_MARKER_OFFSET_CHECK_WA \
	CPK_FENCE_ALL_DBG CPK_CQP_HANG_ON_SUSPENDED_QPS_WA \
	PUSH_INDEX_CHECK_WA MULTICAST_EXTERNAL_PORT_WA

STRIP_FLAGS_KERNEL = UPSTREAM_RELEASE REM_ENDPOINT_TRACKING_UNSUPPORTED IIDC_UPSTREAM

STRIP_FLAGS_C0 = CPK_FPK_ORRQ_FULL MULTICAST_EXTERNAL_PORT_WA

STRIP_FLAGS_DEBUG = DEBUG_SUPPORT

STRIP_FLAGS_CONFIGFS = CONFIGFS_SUPPORT

STRIP_FLAGS_RCA = MEV_RCA

########## build defines ##########
# convenient build settings
# ALL_GEN - combined driver
# ALL_GEN_HVL - combined driver on B0
# SI - Silicon B0 combined driver
# DEBUG - Internal SI combined driver Debug Build
# FPGA - FPGA B0
# KERNEL - Upstream content combined driver
# SIMICS - CPK Simics model
# MEV_SIMICS - MEV Simics model
# #################################
ifeq ($(BUILD),ALL_GEN)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=YES
BUILD_MEV=YES
BUILD_PF_SUPPORT_VF=YES
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=NO
ifeq ($(BUILD_PF_SUPPORT_VF), YES)
SHARED_FILES += $(MEV_SHARED_FILES)
endif
STRIP_FLAGS += -UMEV_HVL -DCPK_INLINE_PERF_ERRATA -DCPK_UP_MAP_WA -DCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -DCPK_PORT_VLAN_WA -DLAG_SUPPORT -DMEV_CRT -URELAX_DB_ALGORITHM -DVCHNL_PRIVILEGED -DVCHNL_UNPRIVILEGED
endif

ifeq ($(BUILD),ALL_GEN_HVL)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=YES
BUILD_MEV=YES
BUILD_PF_SUPPORT_VF=YES
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=YES
ifeq ($(BUILD_PF_SUPPORT_VF), YES)
SHARED_FILES += $(MEV_SHARED_FILES)
endif
STRIP_FLAGS += -DCPK_INLINE_PERF_ERRATA -DCPK_UP_MAP_WA -DCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -DCPK_PORT_VLAN_WA -DLAG_SUPPORT -DMEV_CRT -URELAX_DB_ALGORITHM -DVCHNL_PRIVILEGED -DVCHNL_UNPRIVILEGED
endif

ifeq ($(BUILD),ALL_GEN_HVL_ACC)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=YES
BUILD_MEV=YES
BUILD_FPK=YES
BUILD_PF_SUPPORT_VF=YES
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=YES
ifeq ($(BUILD_PF_SUPPORT_VF), YES)
SHARED_FILES += $(MEV_SHARED_FILES)
endif
STRIP_FLAGS += -DCPK_INLINE_PERF_ERRATA -DCPK_UP_MAP_WA -DCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -DCPK_PORT_VLAN_WA -DLAG_SUPPORT -DMEV_CRT -URELAX_DB_ALGORITHM -DVCHNL_PRIVILEGED -DVCHNL_UNPRIVILEGED
endif

ifeq ($(BUILD),SI)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=YES
BUILD_MEV=NO
BUILD_PF_SUPPORT_VF=YES
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=NO
ifeq ($(BUILD_PF_SUPPORT_VF), YES)
endif
STRIP_FLAGS += -DCPK_INLINE_PERF_ERRATA -DCPK_UP_MAP_WA -DCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -DCPK_PORT_VLAN_WA -DLAG_SUPPORT -UMEV_CRT -DRELAX_DB_ALGORITHM -DVCHNL_PRIVILEGED -DVCHNL_UNPRIVILEGED
endif

ifeq ($(BUILD),DEBUG)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=YES
BUILD_MEV=NO
BUILD_PF_SUPPORT_VF=YES
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=NO
ifeq ($(BUILD_PF_SUPPORT_VF), YES)
endif
STRIP_FLAGS += -DCPK_INLINE_PERF_ERRATA -DCPK_UP_MAP_WA -DCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -DCPK_PORT_VLAN_WA -DDEBUG_SUPPORT_INTERNAL -UMEV_CRT -URELAX_DB_ALGORITHM -DVCHNL_PRIVILEGED -DVCHNL_UNPRIVILEGED
endif

ifeq ($(BUILD),FPGA)
BUILD_FPGA=YES
BUILD_KERNEL=NO
BUILD_B0=YES
BUILD_MEV=NO
BUILD_PF_SUPPORT_VF=NO
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=NO
STRIP_FLAGS += -DCPK_INLINE_PERF_ERRATA -DCPK_UP_MAP_WA -UCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -UMEV_CRT -URELAX_DB_ALGORITHM
endif

ifeq ($(BUILD),KERNEL)
BUILD_FPGA=NO
BUILD_KERNEL=YES
BUILD_B0=NO
BUILD_MEV=NO
BUILD_PF_SUPPORT_VF=NO
BUILD_DEBUG=NO
BUILD_CONFIGFS=NO
BUILD_RCA=NO
filter-list = virtchnl.c virtchnl.h debugfs.c irdma_kcompat.c configfs.c configfs.h \
	      irdma_kcompat.h distro_ver.h rhel_kcompat.h suse_kcompat.h linux_kcompat.h \
	      ubuntu_kcompat.h oracle_kcompat.h
SHARED_FILES := $(filter-out $(filter-list),$(SHARED_FILES))
CORE_FILES := $(filter-out $(filter-list),$(CORE_FILES)) Kconfig.pf
STRIP_FLAGS += -URHEL_7_2 -URHEL_7_7 -URHEL_7_8 -URHEL_7_9 -UDCQCN_CONFIG_SUPPORT -UIB_UMEM_GET_V0 -UIB_UMEM_GET_V1 -UIB_UMEM_GET_V2 -DIB_UMEM_GET_V3 \
	       -UCONFIG_DEV_LIMITS -DCPK_INLINE_PERF_ERRATA -UCPK_UP_MAP_WA -UCPK_READFENCE_MITIGATION -UCPK_DESTROY_QP_FAIL_WA \
	       -USET_BEST_PAGE_SZ_V1 -DSET_BEST_PAGE_SZ_V2 -UIB_IW_PKEY -UIB_IW_MANDATORY_AH_OP -U__OFED_4_8__ -DFIELD_PREP \
	       -UGLOBAL_HDL_LIST -DIIDC_UPSTREAM -UCPK_PORT_VLAN_WA -DIB_DEALLOC_DRIVER_SUPPORT -DRDMA_MMAP_DB_SUPPORT \
	       -Drdma_umem_for_each_dma_block -Dfallthrough -Dibdev_dbg -Drefcount_t -DALLOC_HW_STATS_STRUCT_V2 -ULAG_SUPPORT \
	       -DIB_DEV_CAPS_VER_2 -DNETDEV_TO_IBDEV_SUPPORT -UDUP_CM_NAME_WA -UMEV_CRT -URELAX_DB_ALGORITHM -UVCHNL_PRIVILEGED -UVCHNL_UNPRIVILEGED
	       -DIB_DEV_CAPS_VER_2 -DNETDEV_TO_IBDEV_SUPPORT -DSET_DMABUF -UDUP_CM_NAME_WA -UMEV_CRT -URELAX_DB_ALGORITHM -UVCHNL_PRIVILEGED \
	       -UVCHNL_UNPRIVILEGED -UCVL_VCHNL_UNPRIVILEGED
endif

ifeq ($(BUILD),SIMICS)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=NO
BUILD_MEV=NO
BUILD_PF_SUPPORT_VF=NO
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=NO
STRIP_FLAGS += -DCPK_INLINE_PERF_ERRATA -UCPK_UP_MAP_WA -UCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -UMEV_CRT -URELAX_DB_ALGORITHM
endif

ifeq ($(BUILD),MEV_SIMICS)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=NO
BUILD_MEV=YES
BUILD_PF_SUPPORT_VF=YES
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=YES
SHARED_FILES += $(MEV_SHARED_FILES)
STRIP_FLAGS += -UCPK_INLINE_PERF_ERRATA -UCPK_UP_MAP_WA -UCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -UCPK_PORT_VLAN_WA -ULAG_SUPPORT -DMEV_CRT -URELAX_DB_ALGORITHM -DVCHNL_PRIVILEGED -DVCHNL_UNPRIVILEGED
endif

ifeq ($(BUILD),MEV_VELOCE)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=NO
BUILD_MEV=YES
BUILD_PF_SUPPORT_VF=YES
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=YES
SHARED_FILES += $(MEV_SHARED_FILES)
STRIP_FLAGS += -UCPK_INLINE_PERF_ERRATA -UCPK_UP_MAP_WA -UCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -UCPK_PORT_VLAN_WA -ULAG_SUPPORT -DMEV_CRT -URELAX_DB_ALGORITHM -DVCHNL_PRIVILEGED -DVCHNL_UNPRIVILEGED
endif

ifeq ($(BUILD),MEV_SILICON)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=NO
BUILD_MEV=YES
BUILD_PF_SUPPORT_VF=YES
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=YES
SHARED_FILES += $(MEV_SHARED_FILES)
STRIP_FLAGS += -UCPK_INLINE_PERF_ERRATA -UCPK_UP_MAP_WA -UCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -UCPK_PORT_VLAN_WA -ULAG_SUPPORT -DMEV_CRT -URELAX_DB_ALGORITHM -DVCHNL_PRIVILEGED -DVCHNL_UNPRIVILEGED
endif

ifeq ($(BUILD),MEV_ACC_VELOCE)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=NO
BUILD_MEV=YES
BUILD_PF_SUPPORT_VF=YES
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=YES
SHARED_FILES += $(MEV_SHARED_FILES)
STRIP_FLAGS += -UCPK_INLINE_PERF_ERRATA -UCPK_UP_MAP_WA -UCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -UCPK_PORT_VLAN_WA -ULAG_SUPPORT -DMEV_CRT -URELAX_DB_ALGORITHM -DVCHNL_PRIVILEGED -DVCHNL_UNPRIVILEGED
endif

ifeq ($(BUILD),MEV_ACC_SILICON)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=NO
BUILD_MEV=YES
BUILD_PF_SUPPORT_VF=YES
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=YES
SHARED_FILES += $(MEV_SHARED_FILES)
STRIP_FLAGS += -UCPK_INLINE_PERF_ERRATA -UCPK_UP_MAP_WA -UCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -UCPK_PORT_VLAN_WA -ULAG_SUPPORT -DMEV_CRT -URELAX_DB_ALGORITHM -DVCHNL_PRIVILEGED -DVCHNL_UNPRIVILEGED
endif

ifeq ($(BUILD),MEV_ACC_SIMICS)
BUILD_FPGA=NO
BUILD_KERNEL=NO
BUILD_B0=NO
BUILD_MEV=YES
BUILD_PF_SUPPORT_VF=YES
BUILD_DEBUG=YES
BUILD_CONFIGFS=YES
BUILD_RCA=YES
SHARED_FILES += $(MEV_SHARED_FILES)
STRIP_FLAGS += -UCPK_INLINE_PERF_ERRATA -UCPK_UP_MAP_WA -UCPK_DESTROY_QP_FAIL_WA -DGLOBAL_HDL_LIST -UIIDC_UPSTREAM -UCPK_PORT_VLAN_WA -ULAG_SUPPORT -DMEV_CRT -URELAX_DB_ALGORITHM -DVCHNL_PRIVILEGED -DVCHNL_UNPRIVILEGED
endif

ifeq ($(V),)
V = 0
endif

# pass BUILD_FPGA=YES to include support for FPGA
ifeq ($(BUILD_FPGA),)
BUILD_FPGA=NO
endif

# pass BUILD_KERNEL=YES to include support for KERNEL
ifeq ($(BUILD_KERNEL),)
BUILD_KERNEL=NO
endif

ifeq ($(V),1)
DEBUG = _
endif

define SF
ifeq ($(BUILD_$(1)), YES)
STRIP_FLAGS += $(foreach f, $(STRIP_FLAGS_$(1)), -D$(f))
else
STRIP_FLAGS += $(foreach f, $(STRIP_FLAGS_$(1)), -U$(f))
endif
endef

$(eval $(call SF,FPGA))
$(eval $(call SF,B0))
$(eval $(call SF,KERNEL))
$(eval $(call SF,MEV))
$(eval $(call SF,PF_SUPPORT_VF))
$(eval $(call SF,DEBUG))
$(eval $(call SF,CONFIGFS))
$(eval $(call SF,RCA))

KEEP_FLAGS += $(KEEP_FLAGS_EXTRA)
STRIP_FLAGS += $(STRIP_FLAGS_EXTRA)

# compress whitespace around stripped sections
UNIFDEF_FLAGS = -B

# tools
SPDX    = GPL-2.0 or Linux-OpenIB
UNIFDEF = ../scripts/unifdef
STRIP   = $(UNIFDEF) $(UNIFDEF_FLAGS) -L -t
KEEP    = perl ../scripts/cppkeep.pl "$(STRIP)"
COMMENT = ../scripts/comment_block.sh
PREPROCDEF = ../scripts/codeprep.sed \
	     ../scripts/irdma_print.cocci \

default: all

$(UNIFDEF):
	@$(CC) -o ../scripts/unifdef ../scripts/unifdef.c

clean:
	@rm -rf THESE_UNKNOWN_FLAGS* *.gz \
		$(PF_DRIVER_DIR) *.ko

prep-dirs:
	-@mkdir -p $(PF_DRIVER_DIR)

Makefiles: COMMENTFLAG=-S
Makefiles: UNIFDEF_FLAGS=
Makefiles: $(UNIFDEF) $(PF_DRIVER_DIR)/Kbuild $(PF_DRIVER_DIR)/build.sh


# NOTE: strip flags must always apply first!
# $(1) is source dir, $(2) is source filename, $(3) is dest dir,
# $(4) is extra strip flags
define codeprepare
	@$(COMMENT) $(1)/$(2) \
	    | cat - $(1)/$(2) \
	    | $(STRIP) $(STRIP_FLAGS) $(4) \
	    | $(KEEP) $(KEEP_FLAGS) \
	    | sed -e 's/@VERSION_MAJOR@/${PF_VERSION_MAJOR}/' \
		  -e 's/@VERSION_MINOR@/${PF_VERSION_MINOR}/' \
		  -e 's/@VERSION_BUILD@/${PF_VERSION_BUILD}/' \
		  -e 's/@SPDX@/${SPDX}/g' \
		  -f ../scripts/codeprep.sed \
	    > $(3)/$(2); \
	spatch --sp-file ../scripts/irdma_print.cocci -j 8 --very-quiet --no-show-diff --in-place --include-headers $(3)/$(2);
endef

# NOTE: strip flags must always apply first!
# this one does not add copyright
define codeprepare_nocomment
	@cat $(1)/$(2) \
	    | $(STRIP) $(STRIP_FLAGS) $(4) \
	    | $(KEEP) $(KEEP_FLAGS) \
	    | sed -e 's/@VERSION_MAJOR@/${PF_VERSION_MAJOR}/' \
		  -e 's/@VERSION_MINOR@/${PF_VERSION_MINOR}/' \
		  -e 's/@VERSION_BUILD@/${PF_VERSION_BUILD}/' \
		  -e 's/@SPDX@/${SPDX}/g' \
		  -f ../scripts/codeprep.sed \
	    > $(3)/$(2);
endef

define codeprepare_
	echo STRIP_FLAGS=$(STRIP_FLAGS) $(4) 1>&2;
	echo KEEP_FLAGS=$(KEEP_FLAGS) 1>&2;
	$(call codeprepare,$(1),$(2),$(3),$(4))
endef

version:
	@echo "${PF_VERSION}"

pf-dir: COMMENTFLAG=-C
pf-dir: $(addprefix $(PF_DRIVER_DIR)/,$(CORE_FILES)) \
	$(addprefix $(PF_DRIVER_DIR)/,$(CORE_FILES_NOCOMMENT)) \
	$(addprefix $(PF_DRIVER_DIR)/,$(SHARED_FILES))

define CODEPREP
$(PF_DRIVER_DIR)/$(2): $(1)/$(2) $(UNIFDEF) $(MAKEFILE_LIST) $(PREPROCDEF)
	@echo processing $(1)/$(2) into $(PF_DRIVER_DIR)/$(2)
	$(call codeprepare$(DEBUG),$(1),$(2),$(PF_DRIVER_DIR),$(STRIP_FLAGS_PF))
endef

define CODEPREP_NC
$(PF_DRIVER_DIR)/$(2): $(1)/$(2) $(UNIFDEF) $(MAKEFILE_LIST) $(PREPROCDEF)
	@echo processing $(1)/$(2) into $(PF_DRIVER_DIR)/$(2)
	$(call codeprepare_nocomment,$(1),$(2),$(PF_DRIVER_DIR),$(STRIP_FLAGS_PF))
endef

$(foreach f, Kbuild build.sh, $(eval $(call CODEPREP,CORE,$(f))))
$(foreach f, $(CORE_FILES), $(eval $(call CODEPREP,CORE,$(f))))
$(foreach f, $(SHARED_FILES), $(eval $(call CODEPREP,SHARED,$(f))))
$(foreach f, $(CORE_FILES_NOCOMMENT), $(eval $(call CODEPREP_NC,CORE,$(f))))

all: prep-dirs Makefiles pf-dir $(UNIFDEF)
	@if [ -e THESE_UNKNOWN_FLAGS_NEED_PLACING_IN_BUILD_MK ]; then \
		echo ""; \
		echo "**************************************************"; \
		echo "* WARNING!!  Unknown compile flags in your code! *"; \
		echo "*     Check THESE_UNKNOWN_FLAGS... for the list. *"; \
		echo "**************************************************"; \
		echo ""; \
		cat "THESE_UNKNOWN_FLAGS_NEED_PLACING_IN_BUILD_MK" ; \
		echo ""; \
	fi

#doxygen: Doxyfile all
#	for d in $(PF_DRIVER_DIR) ; do \
#		cd $$d ;                                \
#		doxygen ../Doxyfile ;                   \
#		cd .. ;                                 \
#	done

.PHONY: version prep-dirs Makefiles pf-dir clean
