// SPDX-License-Identifier: GPL-2.0 or Linux-OpenIB
/* Copyright (c) 2018 - 2023 Intel Corporation */
#include "osdep.h"
#include "type.h"
#include "status.h"
#include "protos.h"
#include "ig3rdma_regs.h"
#include "ig3rdma_regs_apf.h"
#include "ig3rdma_regs_avf.h"
#include "ig3rdma_hw.h"

#define IG3_MAX_APFS 1
#define IG3_MAX_AVFS 0

#define IG3_CQPSQ_STAG_PDID_S	46
#define IG3_CQPSQ_STAG_PDID_M	(0x3ffffULL << IG3_CQPSQ_STAG_PDID_S)
#define IG3_CQPSQ_CQ_CQID_S	0
#define IG3_CQPSQ_CQ_CQID_M	(0x3fffffULL << IG3_CQPSQ_CQ_CQID_S)
#define IG3_CQPSQ_CQ_CEQID_S	22
#define IG3_CQPSQ_CQ_CEQID_M	(0x3ffULL << IG3_CQPSQ_CQ_CEQID_S)

#define IG3_COMMIT_FPM_CQCNT_S	0
#define IG3_COMMIT_FPM_CQCNT_M	(0xfffffULL << IG3_COMMIT_FPM_CQCNT_S)
#define IG3_CQPSQ_UPESD_HMCFNID_S 0
#define IG3_CQPSQ_UPESD_HMCFNID_M (0x7fULL << IG3_CQPSQ_UPESD_HMCFNID_S)

#define IG3_RDMA_PE_ADMIN_REQ_IF_VER 0

#define IG3_CPUW_GLPE_IMCREQ_PE_IMCREQ_OP_S 0
#define IG3_CPUW_GLPE_IMCREQ_PE_IMCREQ_OP GENMASK_ULL(15, 0)

#define IG3_CPUW_GLPE_IMCREQ_PE_IMCREQ_IF_VER_S 16
#define IG3_CPUW_GLPE_IMCREQ_PE_IMCREQ_IF_VER GENMASK_ULL(23, 16)

#define IG3_CPUW_GLPE_IMCREQ_PE_IMCREQ_VALID_S 31
#define IG3_CPUW_GLPE_IMCREQ_PE_IMCREQ_VALID BIT_ULL(31)

#define IG3_CPUW_GLPE_IMCRESP_VALID_S 31
#define IG3_CPUW_GLPE_IMCRESP_VALID BIT_ULL(31)

#define IG3_CPUW_GLPE_IMCRESP_ATOMICS_EN_S 0
#define IG3_CPUW_GLPE_IMCRESP_ATOMICS_EN BIT_ULL(0)

#define IG3_CPUW_GLPE_IMCRESP_PF_NUM_S 0
#define IG3_CPUW_GLPE_IMCRESP_PF_NUM GENMASK_ULL(3, 0)

#define IG3_CPUW_GLPE_IMCRESP_HOST_ID_S 4
#define IG3_CPUW_GLPE_IMCRESP_HOST_ID GENMASK_ULL(6, 4)

#define IG3_CPUW_GLPE_IMCRESP_MAX_VSIS_S 16
#define IG3_CPUW_GLPE_IMCRESP_MAX_VSIS GENMASK_ULL(31, 16)

#define IG3_CPUW_GLPE_IMCRESP_LOCALMEM_OFF_S 6
#define IG3_CPUW_GLPE_IMCRESP_LOCALMEM_OFF GENMASK_ULL(31, 6)

struct ig3_rdma_reg {
	u32 reg_offset;
	bool privileged;
};

static struct ig3_rdma_reg ig3rdma_regs[IRDMA_MAX_REGS] = {
	{IG3_APF_CQPTAIL, false},
	{IG3_APF_DBL_CQPDB, false},
	{IG3_APF_CCQPSTATUS, false},
	{IG3_APF_CCQPHIGH, false},
	{IG3_APF_CCQPLOW, false},
	{IG3_APF_DBL_CQARM, false},
	{IG3_APF_DBL_CQACK, false},
	{IG3_APF_DBL_AEQALLOC, false},
	{IG3_APF_CQPERRCODES, false},
	{IG3_APF_DBL_WQEALLOC, false},
	{IG3_GLINT_DYN_CTL(0), false},
	{IG3_DB_ADDR_OFFSET, false },
	/* APF/AVF use from here up only */
	{IG3_GLPCI_LBARCTRL, true},
	{IG3_CPUW_GLPE_CPUSTATUS0, true},
	{IG3_CPUW_GLPE_CPUSTATUS1, true},
	{IG3_CPUW_GLPE_CPUSTATUS2, true},
	{0xFFFFFFFF, true},
	{0xFFFFFFFF, true},
	{0xFFFFFFFF, true},
	{0xFFFFFFFF, true},
	{0xFFFFFFFF, true},
	{0xFFFFFFFF, true},
	{0xFFFFFFFF, true}
};

static u64 ig3rdma_masks[IRDMA_MAX_MASKS] = {
	IG3_APF_CCQPSTATUS_CCQP_DONE_M,
	IG3_APF_CCQPSTATUS_CCQP_ERR_M,
	IG3_CQPSQ_STAG_PDID_M,
	IG3_CQPSQ_CQ_CEQID_M,
	IG3_CQPSQ_CQ_CQID_M,
	IG3_COMMIT_FPM_CQCNT_M,
	IG3_APF_CCQPEXTSTATUS_DBPT_INDEX_M,
	IG3_CQPSQ_UPESD_HMCFNID_M
};

static u8 ig3rdma_shifts[IRDMA_MAX_SHIFTS] = {
	IG3_APF_CCQPSTATUS_CCQP_DONE_S,
	IG3_APF_CCQPSTATUS_CCQP_ERR_S,
	IG3_CQPSQ_STAG_PDID_S,
	IG3_CQPSQ_CQ_CEQID_S,
	IG3_CQPSQ_CQ_CQID_S,
	IG3_COMMIT_FPM_CQCNT_S,
	IG3_APF_CCQPEXTSTATUS_DBPT_INDEX_S,
	IG3_CQPSQ_UPESD_HMCFNID_S
};

/**
 * ig3rdma_ena_irq - Enable interrupt
 * @dev: pointer to the device structure
 * @idx: vector index
 */
static void ig3rdma_ena_irq(struct irdma_sc_dev *dev, u32 idx)
{
	u32 val;
	u32 int_stride = 1; /* one u32 per register*/

	if (dev->is_pf)
		int_stride = 0x400;
	else
		idx -= 1; /* VFs use DYN_CTL_N*/

	val = IG3_GLINT_DYN_CTL_INTENA_M | IG3_GLINT_DYN_CTL_CLEARPBA_M;
	writel(val, dev->hw_regs[IRDMA_GLINT_DYN_CTL] + (idx * int_stride));
}

/**
 * ig3rdma_disable_irq - Disable interrupt
 * @dev: pointer to the device structure
 * @idx: vector index
 */
static void ig3rdma_disable_irq(struct irdma_sc_dev *dev, u32 idx)
{
	u32 int_stride = 1; /* one u32 per register*/

	if (dev->is_pf)
		int_stride = 0x400;
	else
		idx -= 1; /* VFs use DYN_CTL_N*/

	writel(0, dev->hw_regs[IRDMA_GLINT_DYN_CTL] + (idx * int_stride));
}

/**
 * ig3rdma_cfg_ceq - Configure CEQ interrupt
 * @dev: pointer to the device structure
 * @ceq_id: Completion Event Queue ID
 * @idx: vector index
 * @enable: enable the ceq vector if set, else disable
 */
static void ig3rdma_cfg_ceq(struct irdma_sc_dev *dev, u32 ceq_id, u32 idx,
			    bool enable)
{
	int status_code;
	struct irdma_set_interrupt_info setintinfo = {};

	setintinfo.ceq_agent_interrupt_index = idx;
	setintinfo.ceq_id = ceq_id;
	setintinfo.set_ceq_int = enable;
	setintinfo.enable_int = 1;
	status_code = irdma_cqp_set_interrupt_cmd(dev, &setintinfo);
	if (status_code)
		ibdev_dbg(to_ibdev(dev), "ERR: IRDMA_ERR_CFG_CEQ");
}

/**
 * ig3rdma_cfg_aeq - Configure AEQ interrupt
 * @dev: pointer to the device structure
 * @idx: vector index
 * @enable: enable the aeq vector if set, else disable
 */
static void ig3rdma_cfg_aeq(struct irdma_sc_dev *dev, u32 idx, bool enable)
{
	struct irdma_set_interrupt_info setintinfo = {};
	int status_code;

	setintinfo.aeq_agent_interrupt_index = idx;
	setintinfo.set_aeq_int = enable;
	setintinfo.enable_int = 1;
	status_code = irdma_cqp_set_interrupt_cmd(dev, &setintinfo);
	if (status_code)
		ibdev_dbg(to_ibdev(dev), "ERR: IRDMA_ERR_CFG_AEQ");
}

static const struct irdma_irq_ops ig3rdma_irq_ops = {
	.irdma_cfg_aeq = ig3rdma_cfg_aeq,
	.irdma_cfg_ceq = ig3rdma_cfg_ceq,
	.irdma_dis_irq = ig3rdma_disable_irq,
	.irdma_en_irq = ig3rdma_ena_irq,
};

/**
 * ig3rdma_post_cqp_create - additional processing after CQP create
 * @dev: device structure
 */
int ig3rdma_post_cqp_create(struct irdma_sc_dev *dev)
{
	/* map CQP DB page */
	dev->db_addr = irdma_ioremap_db_pg(dev, 0);
	if (!dev->db_addr)
		return -EOPNOTSUPP;
	if (dev->privileged) {
		/* set up doorbell variables using mapped DB page */
		dev->wqe_alloc_db =
			(u32 __iomem *)(dev->hw->hw_addr +
				      ig3rdma_regs[IRDMA_WQEALLOC].reg_offset);
		dev->cq_arm_db =
			(u32 __iomem *)(dev->hw->hw_addr +
				      ig3rdma_regs[IRDMA_CQARM].reg_offset);
		dev->aeq_alloc_db =
			(u32 __iomem *)(dev->hw->hw_addr +
				      ig3rdma_regs[IRDMA_AEQALLOC].reg_offset);
		dev->cqp_db =
			(u32 __iomem *)(dev->hw->hw_addr +
				      ig3rdma_regs[IRDMA_CQPDB].reg_offset);
		dev->cq_ack_db =
			(u32 __iomem *)(dev->hw->hw_addr +
			  	      ig3rdma_regs[IRDMA_CQACK].reg_offset);
		dev->ceq_itr_mask_db = NULL; /* TODO */
		dev->aeq_itr_mask_db = NULL; /* TODO */
	}
	return 0;
}

/**
 * ig3rdma_post_cqp_destroy - additional processing after CQP destroy
 * @dev: device structure
 */
void ig3rdma_post_cqp_destroy(struct irdma_sc_dev *dev)
{
	/* unset doorbell registers */
	dev->wqe_alloc_db = NULL;
	dev->cq_arm_db = NULL;
	dev->aeq_alloc_db = NULL;
	dev->cqp_db = NULL;
	dev->cq_ack_db = NULL;
	dev->ceq_itr_mask_db = NULL; /* TODO */
	dev->aeq_itr_mask_db = NULL; /* TODO */

	/* unmap CQP DB page */
	irdma_iounmap(dev);
}

static const struct irdma_hw_stat_map ig3rdma_hw_stat_map[] = {
	[IRDMA_HW_STAT_INDEX_RXVLANERR] =	{   0, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4RXOCTS] =	{   8, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4RXPKTS] =	{  16, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4RXDISCARD] =	{  24, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4RXTRUNC] =	{  32, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4RXFRAGS] =	{  40, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4RXMCOCTS] =	{  48, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4RXMCPKTS] =	{  56, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6RXOCTS] =	{  64, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6RXPKTS] =	{  72, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6RXDISCARD] =	{  80, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6RXTRUNC] =	{  88, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6RXFRAGS] =	{  96, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6RXMCOCTS] =	{ 104, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6RXMCPKTS] =	{ 112, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4TXOCTS] =	{ 120, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4TXPKTS] =	{ 128, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4TXFRAGS] =	{ 136, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4TXMCOCTS] =	{ 144, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4TXMCPKTS] =	{ 152, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6TXOCTS] =	{ 160, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6TXPKTS] =	{ 168, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6TXFRAGS] =	{ 176, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6TXMCOCTS] =	{ 184, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6TXMCPKTS] =	{ 192, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP4TXNOROUTE] =	{ 200, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_IP6TXNOROUTE] =	{ 208, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_TCPRTXSEG] =	{ 216, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_TCPRXOPTERR] =	{ 224, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_TCPRXPROTOERR] =	{ 232, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_TCPTXSEG] =	{ 240, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_TCPRXSEGS] =	{ 248, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_UDPRXPKTS] =	{ 256, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_UDPTXPKTS] =	{ 264, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMARXWRS] =	{ 272, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMARXRDS] =	{ 280, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMARXSNDS] =	{ 288, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMATXWRS] =	{ 296, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMATXRDS] =	{ 304, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMATXSNDS] =	{ 312, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMAVBND] =	{ 320, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMAVINV] =	{ 328, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RXNPECNMARKEDPKTS] = { 336, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RXRPCNPHANDLED] =	{ 344, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RXRPCNPIGNORED] =	{ 352, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_TXNPCNPSENT] =	{ 360, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RNR_SENT] =	{ 368, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RNR_RCVD] =	{ 376, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMAORDLMTCNT] =	{ 384, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMAIRDLMTCNT] =	{ 392, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMARXATS] =	{ 408, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RDMATXATS] =	{ 416, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_NAKSEQERR] =	{ 424, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_NAKSEQERR_IMPLIED] = { 432, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RTO] =		{ 440, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_RXOOOPKTS] =	{ 448, 0, IRDMA_MAX_STATS_64 },
	[IRDMA_HW_STAT_INDEX_ICRCERR] =		{ 456, 0, IRDMA_MAX_STATS_64 },
};

static const struct irdma_vsi_ops ig3_vsi_ops = {
	.vsi_update_stats = sc_vsi_update_stats,
};

void ig3rdma_init_hw(struct irdma_sc_dev *dev)
{
	int i;
	u8 __iomem *hw_addr;

	for (i = 0; i < IRDMA_MAX_REGS; ++i) {
		if (ig3rdma_regs[i].privileged)
			hw_addr = dev->hw->priv_hw_addr;
		else
			hw_addr = dev->hw->hw_addr;

		if (i == IRDMA_DB_ADDR_OFFSET)
			hw_addr = NULL;

		dev->hw_regs[i] =
			(u32 __iomem *)(hw_addr + ig3rdma_regs[i].reg_offset);
	}

	for (i = 0; i < IRDMA_MAX_SHIFTS; ++i)
		dev->hw_shifts[i] = ig3rdma_shifts[i];

	for (i = 0; i < IRDMA_MAX_MASKS; ++i)
		dev->hw_masks[i] = ig3rdma_masks[i];

	dev->irq_ops = &ig3rdma_irq_ops;
	dev->vsi_ops = &ig3_vsi_ops;
	dev->hw_stats_map = ig3rdma_hw_stat_map;

	dev->hw_attrs.uk_attrs.hw_rev = IRDMA_GEN_3;
	dev->hw_attrs.uk_attrs.max_hw_wq_frags = IG3RDMA_MAX_WQ_FRAGMENT_COUNT;
	dev->hw_attrs.uk_attrs.max_hw_read_sges = IG3RDMA_MAX_SGE_RD;
	dev->hw_attrs.uk_attrs.max_hw_sq_chunk = IRDMA_MAX_QUANTA_PER_WR;
	dev->hw_attrs.first_hw_vf_fpm_id = 0;
	dev->hw_attrs.max_hw_vf_fpm_id = IG3_MAX_APFS + IG3_MAX_AVFS;
	dev->hw_attrs.uk_attrs.feature_flags |= IRDMA_FEATURE_ATOMIC_OPS;
	dev->hw_attrs.uk_attrs.feature_flags |= IRDMA_FEATURE_64_BYTE_CQE;
	dev->hw_attrs.uk_attrs.feature_flags |= IRDMA_FEATURE_CQE_TIMESTAMPING;

	/* Currently only affects user space */
	dev->hw_attrs.uk_attrs.feature_flags |= IRDMA_FEATURE_ENFORCE_SQ_SIZE;
	dev->hw_attrs.uk_attrs.feature_flags |= IRDMA_FEATURE_SRQ;
	dev->hw_attrs.uk_attrs.feature_flags |= IRDMA_FEATURE_RTS_AE |
						IRDMA_FEATURE_CQ_RESIZE;
	dev->hw_attrs.page_size_cap = SZ_4K | SZ_2M | SZ_1G;
	dev->hw_attrs.max_hw_ird = IG3RDMA_MAX_IRD_SIZE;
	dev->hw_attrs.max_hw_ord = IG3RDMA_MAX_ORD_SIZE;
	dev->hw_attrs.max_stat_inst = IG3RDMA_MAX_STATS_COUNT;
	dev->hw_attrs.max_stat_idx = IRDMA_HW_STAT_INDEX_MAX_GEN_3;
	if (dev->hw_wa & REDUCE_ORD_IRD) {
		dev->hw_attrs.max_hw_ird = 128;
		dev->hw_attrs.max_hw_ord = 128;
	}
	if (dev->hw_wa & REDUCE_ORD_IRD_SIMICS) {
		dev->hw_attrs.max_hw_ird = 128;
		dev->hw_attrs.max_hw_ord = 128;
	}
	if (dev->hw_wa & CQP_COMPL_WAIT)
		dev->hw_attrs.max_cqp_compl_wait_time_ms *= 50;
	dev->hw_attrs.uk_attrs.min_hw_wq_size = IG3RDMA_MIN_WQ_SIZE;
	dev->hw_attrs.uk_attrs.max_hw_srq_quanta = IRDMA_SRQ_MAX_QUANTA;
}
